# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import unittest

import processscheduler as ps

# in this test, two tasks. The start of the first one and the end of the second
# one are constrained by a linear function
# the maximum of task_1.end is 20 (in this case task_1.start is 0)
# the maximum of task_2.end is 20 (in this case task_2.start is 0)
# what happens if we look for the maximum of both task_1 and task_2 ends ?


class MultiObjective(unittest.TestCase):
    def test_multi_two_tasks_1(self) -> None:
        pb = ps.SchedulingProblem("MultiObjective1", horizon=20)
        task_1 = ps.FixedDurationTask("task1", duration=3)
        task_2 = ps.FixedDurationTask("task2", duration=3)

        pb.add_constraint(task_1.end == 20 - task_2.start)

        # Maximize only task_1 end
        ind = ps.Indicator("Task1End", task_1.end)
        pb.maximize_indicator(ind)

        solution = ps.SchedulingSolver(pb).solve()

        self.assertTrue(solution)
        self.assertEqual(solution.tasks[task_1.name].end, 20)
        self.assertEqual(solution.tasks[task_2.name].start, 0)

    def test_multi_two_tasks_lex(self) -> None:
        # the same model, optimize task2 end
        pb = ps.SchedulingProblem("MultiObjective2", horizon=20)
        task_1 = ps.FixedDurationTask("task1", duration=3)
        task_2 = ps.FixedDurationTask("task2", duration=3)

        pb.add_constraint(task_1.end == 20 - task_2.start)

        # Maximize only task_2 end
        ind = ps.Indicator("Task2End", task_2.end)
        pb.maximize_indicator(ind)

        solution = ps.SchedulingSolver(pb).solve()

        self.assertTrue(solution)
        self.assertEqual(solution.tasks[task_1.name].start, 0)
        self.assertEqual(solution.tasks[task_2.name].end, 20)


if __name__ == "__main__":
    unittest.main()



# ProcessScheduler benchmark
import sys
import time
from datetime import datetime
import subprocess
import os
import platform
import psutil
import uuid

import matplotlib.pyplot as plt
import processscheduler as ps
import z3

#
# Display machine identification
#
def get_size(byt, suffix="B"):
    """
    Scale bytes to its proper format
    e.g:
        1253656 => '1.20MB'
        1253656678 => '1.17GB'
    """
    # Code from https://www.thepythoncode.com/article/get-hardware-system-information-python
    factor = 1024
    for unit in ["", "K", "M", "G", "T", "P"]:
        if byt < factor:
            return f"{byt:.2f}{unit}{suffix}"
        byt /= factor


bench_id = uuid.uuid4().hex
bench_date = datetime.now()
print("#### Benchmark information header ####")
print("Date:", bench_date)
print("Id:", bench_id)
print("Software:")
print("\tPython version:", platform.python_version())
print("\tProcessScheduler version:", ps.__VERSION__)
commit_short_hash = subprocess.check_output(
    ["git", "rev-parse", "--short", "HEAD"]
).strip()
print("\tz3 version:", z3.Z3_get_full_version())

print("\tProcessScheduler commit number:", commit_short_hash.decode("utf-8"))
os_info = os.uname()
print("OS:")
print("\tOS:", os_info.sysname)
print("\tOS Release:", os_info.release)
print("\tOS Version:", os_info.version)
print("Hardware:")
print("\tMachine:", os_info.machine)
print("\tPhysical cores:", psutil.cpu_count(logical=False))
print("\tTotal cores:", psutil.cpu_count(logical=True))
# CPU frequencies
cpufreq = psutil.cpu_freq()
print(f"\tMax Frequency: {cpufreq.max:.2f}Mhz")
print(f"\tMin Frequency: {cpufreq.min:.2f}Mhz")
# get the memory details
svmem = psutil.virtual_memory()
print(f"\tTotal memory: {get_size(svmem.total)}")

model_creation_times = []
computation_times = []

test_init_time = time.perf_counter()
all_logics = [
    "QF_LRA",
    "HORN",
    "QF_LIA",
    "QF_RDL",
    "QF_IDL",
    "QF_AUFLIA",
    "QF_ALIA",
    "QF_AUFLIRA",
    "QF_AUFNIA",
    "QF_AUFNIRA",
    "QF_ANIA",
    "QF_LIRA",
    "QF_UFLIA",
    "QF_UFLRA",
    "QF_UFIDL",
    "QF_UFRDL",
    "QF_NIRA",
    "QF_UFNRA",
    "QF_UFNIA",
    "QF_UFNIRA",
    "QF_S",
    "QF_SLIA",
    "UFIDL",
    "HORN",
    "QF_FPLRA",
]
# skipped: "QF_NIA",


num_dev_teams = 40
print("-> Num dev teams:", num_dev_teams)
# Teams and Resources
num_resource_a = 2
num_resource_b = 2

init_time = time.perf_counter()
# Resources
def get_problem():
    digital_transformation = ps.SchedulingProblem("DigitalTransformation")
    print("Create model...", end="")
    r_a = [ps.Worker("A_%i" % (i + 1)) for i in range(num_resource_a)]
    r_b = [ps.Worker("B_%i" % (i + 1)) for i in range(num_resource_b)]

    # Dev Team Tasks
    # For each dev_team pick one resource a and one resource b.
    ts_team_migration = [
        ps.FixedDurationTask("DevTeam_%i" % (i + 1), duration=1, priority=i % 3 + 1)
        for i in range(num_dev_teams)
    ]
    for t_team_migration in ts_team_migration:
        t_team_migration.add_required_resource(ps.SelectWorkers(r_a))
        t_team_migration.add_required_resource(ps.SelectWorkers(r_b))

    # solve
    digital_transformation.add_objective_priorities()
    digital_transformation.add_objective_flowtime()
    # digital_transformation.add_objective_makespan(weight=10)
    return digital_transformation


top1 = time.perf_counter()
mt = 60
results = {}
for logics in all_logics:
    digital_transformation = get_problem()
    top2 = time.perf_counter()
    solver = ps.SchedulingSolver(
        digital_transformation,
        logics=logics,
        random_seed=False,
        parallel=False,
        max_time=20,
    )
    solution = solver.solve()
    if solution:
        flowtime_result = solution.indicators["FlowTime"]
        priority_result = solution.indicators["PriorityTotal"]
    else:
        flowtime_result, priority_result = None, None
    computing_time = time.perf_counter() - top2

    computation_times.append(computing_time)

    print("Logics:", logics, "Total Time:", computing_time)
    results[logics] = (computing_time, flowtime_result, priority_result)
test_final_time = time.perf_counter()
print("TOTAL BENCH TIME:", test_final_time - test_init_time)
print("Results:")
print(results)



# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import unittest

import processscheduler as ps


class TestOptionalTask(unittest.TestCase):
    """For each of these tests, we consider two tests: whether the optional task
    is scheduled or not.
    """

    def test_optional_task_start_at_1(self) -> None:
        """Task can be scheduled."""
        pb = ps.SchedulingProblem("OptionalTaskStartAt1", horizon=6)
        task_1 = ps.FixedDurationTask("task1", duration=3, optional=True)
        pb.add_constraint(ps.TaskStartAt(task_1, 1))

        # Force schedule, otherwise by default it is not scheduled
        pb.add_constraint(task_1.scheduled == True)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertTrue(solution.tasks[task_1.name].scheduled)
        self.assertEqual(solution.tasks[task_1.name].start, 1)
        self.assertEqual(solution.tasks[task_1.name].end, 4)

    def test_optional_task_start_at_2(self) -> None:
        """Task cannot be scheduled."""
        pb = ps.SchedulingProblem("OptionalTaskStartAt2", horizon=2)
        task_1 = ps.FixedDurationTask("task1", duration=3, optional=True)
        pb.add_constraint(ps.TaskStartAt(task_1, 1))

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertFalse(solution.tasks[task_1.name].scheduled)

    def test_optional_task_end_at_1(self) -> None:
        """Task can be scheduled."""
        pb = ps.SchedulingProblem("OptionalTaskEndAt1", horizon=6)
        task_1 = ps.FixedDurationTask("task1", duration=3, optional=True)
        pb.add_constraint(ps.TaskEndAt(task_1, 4))

        # Force schedule, otherwise by default it is not scheduled
        pb.add_constraint(task_1.scheduled == True)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertTrue(solution.tasks[task_1.name].scheduled)
        self.assertEqual(solution.tasks[task_1.name].start, 1)
        self.assertEqual(solution.tasks[task_1.name].end, 4)

    def test_optional_task_end_at_2(self) -> None:
        """Task cannot be scheduled."""
        pb = ps.SchedulingProblem("OptionalTaskEndAt2", horizon=2)
        task_1 = ps.FixedDurationTask("task1", duration=3, optional=True)
        pb.add_constraint(ps.TaskEndAt(task_1, 4))

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertFalse(solution.tasks[task_1.name].scheduled)

    def test_optional_tasks_start_sync_1(self) -> None:
        """Tasks can be scheduled."""
        pb = ps.SchedulingProblem("OptionalTasksStartSynced1", horizon=6)
        task_1 = ps.FixedDurationTask("task1", duration=3)  # mandatory
        task_2 = ps.FixedDurationTask("task2", duration=4, optional=True)  # optional

        pb.add_constraint(ps.TaskStartAt(task_1, 2))
        pb.add_constraint(ps.TasksStartSynced(task_1, task_2))

        # Force schedule, otherwise by default it is not scheduled
        pb.add_constraint(task_2.scheduled == True)

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertTrue(solution.tasks[task_1.name].scheduled)
        self.assertTrue(solution.tasks[task_2.name].scheduled)
        self.assertEqual(solution.tasks[task_1.name].start, 2)
        self.assertEqual(solution.tasks[task_1.name].end, 5)
        self.assertEqual(solution.tasks[task_2.name].start, 2)
        self.assertEqual(solution.tasks[task_2.name].end, 6)

    def test_optional_tasks_start_sync_2(self) -> None:
        """Task 2 cannot be scheduled."""
        pb = ps.SchedulingProblem("OptionalTasksStartSynced2", horizon=5)
        task_1 = ps.FixedDurationTask("task1", duration=3)  # mandatory
        task_2 = ps.FixedDurationTask("task2", duration=4, optional=True)  # optional

        pb.add_constraint(ps.TaskStartAt(task_1, 2))
        pb.add_constraint(ps.TasksStartSynced(task_1, task_2))

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertTrue(solution.tasks[task_1.name].scheduled)
        self.assertFalse(solution.tasks[task_2.name].scheduled)
        self.assertEqual(solution.tasks[task_1.name].start, 2)
        self.assertEqual(solution.tasks[task_1.name].end, 5)

    def test_optional_tasks_end_sync_1(self) -> None:
        """Tasks can be scheduled."""
        pb = ps.SchedulingProblem("OptionalTasksEndSynced1", horizon=6)
        task_1 = ps.FixedDurationTask("task1", duration=3)  # mandatory
        task_2 = ps.FixedDurationTask("task2", duration=4, optional=True)  # optional

        pb.add_constraint(ps.TaskEndAt(task_1, 6))
        pb.add_constraint(ps.TasksEndSynced(task_1, task_2))

        # Force schedule, otherwise by default it is not scheduled
        pb.add_constraint(task_2.scheduled == True)

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertTrue(solution.tasks[task_1.name].scheduled)
        self.assertTrue(solution.tasks[task_2.name].scheduled)
        self.assertEqual(solution.tasks[task_1.name].start, 3)
        self.assertEqual(solution.tasks[task_1.name].end, 6)
        self.assertEqual(solution.tasks[task_2.name].start, 2)
        self.assertEqual(solution.tasks[task_2.name].end, 6)

    def test_optional_tasks_end_sync_2(self) -> None:
        """Task 2 cannot be scheduled."""
        pb = ps.SchedulingProblem("OptionalTasksEndSynced2", horizon=3)
        task_1 = ps.FixedDurationTask("task1", duration=3)  # mandatory
        task_2 = ps.FixedDurationTask("task2", duration=4, optional=True)  # optional

        pb.add_constraint(ps.TaskEndAt(task_1, 3))
        pb.add_constraint(ps.TasksStartSynced(task_1, task_2))

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertTrue(solution.tasks[task_1.name].scheduled)
        self.assertFalse(solution.tasks[task_2.name].scheduled)
        self.assertEqual(solution.tasks[task_1.name].start, 0)
        self.assertEqual(solution.tasks[task_1.name].end, 3)

    def test_optional_tasks_dont_overlap_1(self) -> None:
        """Tasks can be scheduled."""
        pb = ps.SchedulingProblem("OptionalTasksDontOverlap1", horizon=7)
        task_1 = ps.FixedDurationTask("task1", duration=3)  # mandatory
        task_2 = ps.FixedDurationTask("task2", duration=4, optional=True)  # optional

        pb.add_constraint(ps.TaskStartAt(task_1, 0))
        pb.add_constraint(ps.TasksDontOverlap(task_1, task_2))

        # Force schedule, otherwise by default it is not scheduled
        pb.add_constraint(task_2.scheduled == True)

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertTrue(solution.tasks[task_1.name].scheduled)
        self.assertTrue(solution.tasks[task_2.name].scheduled)
        self.assertEqual(solution.tasks[task_1.name].start, 0)
        self.assertEqual(solution.tasks[task_1.name].end, 3)
        self.assertEqual(solution.tasks[task_2.name].start, 3)
        self.assertEqual(solution.tasks[task_2.name].end, 7)

    def test_optional_tasks_dont_overlap_2(self) -> None:
        """Task 2 cannot be scheduled."""
        pb = ps.SchedulingProblem("OptionalTasksDontOverlap2", horizon=3)
        task_1 = ps.FixedDurationTask("task1", duration=3)  # mandatory
        task_2 = ps.FixedDurationTask("task2", duration=4, optional=True)  # optional

        pb.add_constraint(ps.TaskStartAt(task_1, 0))
        pb.add_constraint(ps.TasksDontOverlap(task_1, task_2))

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertTrue(solution.tasks[task_1.name].scheduled)
        self.assertFalse(solution.tasks[task_2.name].scheduled)
        self.assertEqual(solution.tasks[task_1.name].start, 0)
        self.assertEqual(solution.tasks[task_1.name].end, 3)

    def test_optional_tasks_precedence_1(self) -> None:
        """Tasks can be scheduled."""
        pb = ps.SchedulingProblem("OptionalTasksPrecedence1", horizon=9)
        task_1 = ps.FixedDurationTask("task1", duration=3)  # mandatory
        task_2 = ps.FixedDurationTask("task2", duration=4, optional=True)  # optional

        pb.add_constraint(ps.TaskStartAt(task_1, 0))
        pb.add_constraint(ps.TaskPrecedence(task_1, task_2, offset=2))

        # Force schedule, otherwise by default it is not scheduled
        pb.add_constraint(task_2.scheduled == True)

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertTrue(solution.tasks[task_1.name].scheduled)
        self.assertTrue(solution.tasks[task_2.name].scheduled)
        self.assertEqual(solution.tasks[task_1.name].start, 0)
        self.assertEqual(solution.tasks[task_1.name].end, 3)
        self.assertEqual(solution.tasks[task_2.name].start, 5)
        self.assertEqual(solution.tasks[task_2.name].end, 9)

    def test_optional_tasks_precedence_2(self) -> None:
        """Task 2 cannot be scheduled."""
        pb = ps.SchedulingProblem("OptionalTasksPrecedence2", horizon=8)
        task_1 = ps.FixedDurationTask("task1", duration=3)  # mandatory
        task_2 = ps.FixedDurationTask("task2", duration=4, optional=True)  # optional

        pb.add_constraint(ps.TaskStartAt(task_1, 0))
        pb.add_constraint(ps.TaskPrecedence(task_1, task_2, offset=2))

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertTrue(solution.tasks[task_1.name].scheduled)
        self.assertFalse(solution.tasks[task_2.name].scheduled)
        self.assertEqual(solution.tasks[task_1.name].start, 0)
        self.assertEqual(solution.tasks[task_1.name].end, 3)

    def test_optional_tasks_start_after_strict_start_after_lax(self) -> None:
        pb = ps.SchedulingProblem("OptionalTasksStartAfter", horizon=8)
        task_1 = ps.FixedDurationTask("task1", duration=3)  # mandatory
        task_2 = ps.FixedDurationTask("task2", duration=4, optional=True)  # optional

        # force task_2 to be scheduled
        pb.add_constraint(task_2.scheduled == True)

        pb.add_constraint(ps.TaskStartAfterStrict(task_2, 1))
        pb.add_constraint(ps.TaskStartAfterLax(task_2, 4))

        solver = ps.SchedulingSolver(pb, debug=True)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertTrue(solution.tasks[task_1.name].scheduled)
        self.assertTrue(solution.tasks[task_2.name].scheduled)
        self.assertEqual(solution.tasks[task_2.name].start, 4)
        self.assertEqual(solution.tasks[task_2.name].end, 8)

    def test_optional_tasks_end_before_strict_end_before_lax(self) -> None:
        pb = ps.SchedulingProblem("OptionalTasksEndBefore", horizon=8)
        task_1 = ps.FixedDurationTask("task1", duration=3)  # mandatory
        task_2 = ps.FixedDurationTask("task2", duration=4, optional=True)  # optional

        # force task_2 to be scheduled
        pb.add_constraint(task_2.scheduled == True)

        pb.add_constraint(ps.TaskEndBeforeStrict(task_2, 7))
        pb.add_constraint(ps.TaskEndBeforeLax(task_2, 4))

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertTrue(solution.tasks[task_1.name].scheduled)
        self.assertTrue(solution.tasks[task_2.name].scheduled)
        self.assertEqual(solution.tasks[task_2.name].start, 0)
        self.assertEqual(solution.tasks[task_2.name].end, 4)

    def test_optional_condition_1(self) -> None:
        """A task scheduled only if the horizon is > 10."""
        pb = ps.SchedulingProblem("OptionalCondition1")
        task_1 = ps.FixedDurationTask("task1", duration=13)  # mandatory
        task_2 = ps.FixedDurationTask("task2", duration=4, optional=True)  # optional

        cond = ps.OptionalTaskConditionSchedule(task_2, pb.horizon > 10)
        pb.add_constraint(cond)

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertTrue(solution.tasks[task_1.name].scheduled)
        self.assertTrue(solution.tasks[task_2.name].scheduled)

    def test_optional_condition_2(self) -> None:
        """A task scheduled only if the horizon is > 10."""
        pb = ps.SchedulingProblem("OptionalCondition2", horizon=9)
        task_1 = ps.FixedDurationTask("task1", duration=9)  # mandatory
        task_2 = ps.FixedDurationTask("task2", duration=4, optional=True)  # optional

        cond = ps.OptionalTaskConditionSchedule(task_2, pb.horizon > 10)
        pb.add_constraint(cond)

        solver = ps.SchedulingSolver(pb, random_seed=True)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertTrue(solution.tasks[task_1.name].scheduled)
        self.assertFalse(solution.tasks[task_2.name].scheduled)

    def test_optional_task_dependency_1(self) -> None:
        """task_2 is scheduled, because task_1 is."""
        pb = ps.SchedulingProblem("OptionalDependency1")
        task_1 = ps.FixedDurationTask("task1", duration=9)  # mandatory
        task_2 = ps.FixedDurationTask("task2", duration=4, optional=True)  # optional

        cond = ps.OptionalTasksDependency(task_1, task_2)
        pb.add_constraint(cond)

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertTrue(solution.tasks[task_1.name].scheduled)
        self.assertTrue(solution.tasks[task_2.name].scheduled)

    def test_optional_task_dependency_2(self) -> None:
        """task_3 is not scheduled, because task_2 should not."""
        pb = ps.SchedulingProblem("OptionalDependency2", horizon=9)
        task_1 = ps.FixedDurationTask("task1", duration=5)  # mandatory
        task_2 = ps.FixedDurationTask("task2", duration=4, optional=True)  # optional
        task_3 = ps.FixedDurationTask("task3", duration=1, optional=True)  # optional

        cond = ps.OptionalTaskConditionSchedule(task_2, pb.horizon > 10)
        pb.add_constraint(cond)

        dep = ps.OptionalTasksDependency(task_2, task_3)
        pb.add_constraint(dep)

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertTrue(solution.tasks[task_1.name].scheduled)
        self.assertFalse(solution.tasks[task_2.name].scheduled)
        self.assertFalse(solution.tasks[task_3.name].scheduled)

    def test_optional_task_dependency_3(self) -> None:
        """Type checking."""
        pb = ps.SchedulingProblem("OptionalDependency3", horizon=9)
        task_1 = ps.FixedDurationTask("task1", duration=5)  # mandatory
        task_2 = ps.FixedDurationTask("task2", duration=4)  # mandatory
        task_3 = ps.FixedDurationTask("task3", duration=1)  # mandatory

        with self.assertRaises(TypeError):
            ps.OptionalTaskConditionSchedule(task_1, pb.horizon > 10)

        with self.assertRaises(TypeError):
            ps.OptionalTasksDependency(task_2, task_3)

    def test_optional_task_single_worker_1(self) -> None:
        pb = ps.SchedulingProblem("OptionalTaskSingleWorker1", horizon=6)
        task_1 = ps.FixedDurationTask("task1", duration=3, optional=True)
        pb.add_constraint(ps.TaskStartAt(task_1, 1))
        worker_1 = ps.Worker("Worker1")
        task_1.add_required_resource(worker_1)
        # Force schedule, otherwise by default it is not scheduled
        pb.add_constraint(task_1.scheduled == True)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertTrue(solution.tasks[task_1.name].scheduled)

    def test_optional_task_singleworker_2(self) -> None:
        pb = ps.SchedulingProblem("OptionalTaskSingleWorker2", horizon=6)
        task_1 = ps.FixedDurationTask("task1", duration=3, optional=True)
        pb.add_constraint(ps.TaskStartAt(task_1, 1))
        worker_1 = ps.Worker("Worker1")
        task_1.add_required_resource(worker_1)
        # Force schedule to False
        pb.add_constraint(task_1.scheduled == False)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertFalse(solution.tasks[task_1.name].scheduled)

    def test_optional_task_two_workers_1(self) -> None:
        pb = ps.SchedulingProblem("OptionalTaskTwoWorkers1", horizon=6)
        task_1 = ps.FixedDurationTask("task1", duration=3, optional=True)
        pb.add_constraint(ps.TaskStartAt(task_1, 1))
        worker_1 = ps.Worker("Worker1")
        worker_2 = ps.Worker("Worker2")
        task_1.add_required_resources([worker_1, worker_2])
        # Force schedule, otherwise by default it is not scheduled
        pb.add_constraint(task_1.scheduled == True)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertTrue(solution.tasks[task_1.name].scheduled)

    def test_optional_task_two_workers_2(self) -> None:
        pb = ps.SchedulingProblem("OptionalTaskTwoWorkers2")
        task_1 = ps.FixedDurationTask("task1", duration=3, optional=True)

        worker_1 = ps.Worker("Worker1")
        worker_2 = ps.Worker("Worker2")
        task_1.add_required_resources([worker_1, worker_2])
        # Force schedule, otherwise by default it is not scheduled
        pb.add_constraint(task_1.scheduled == False)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertFalse(solution.tasks[task_1.name].scheduled)

    def test_optional_task_select_workers_1(self) -> None:
        pb = ps.SchedulingProblem("OptionalTaskSelectWorkers1")
        task_1 = ps.FixedDurationTask("task1", duration=3, optional=True)
        pb.add_constraint(ps.TaskStartAt(task_1, 1))
        worker_1 = ps.Worker("Worker1")
        worker_2 = ps.Worker("Worker2")
        task_1.add_required_resource(ps.SelectWorkers([worker_1, worker_2], 1))
        # Force schedule, otherwise by default it is not scheduled
        pb.add_constraint(task_1.scheduled == True)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertTrue(solution.tasks[task_1.name].scheduled)
        self.assertEqual(len(solution.tasks[task_1.name].assigned_resources), 1)

    def test_optional_task_select_workers_2(self) -> None:
        pb = ps.SchedulingProblem("OptionalTaskSelectWorkers2")
        task_1 = ps.FixedDurationTask("task1", duration=3, optional=True)
        pb.add_constraint(ps.TaskStartAt(task_1, 1))
        worker_1 = ps.Worker("Worker1")
        worker_2 = ps.Worker("Worker2")
        task_1.add_required_resource(ps.SelectWorkers([worker_1, worker_2], 1))
        # Force schedule False
        pb.add_constraint(task_1.scheduled == False)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertFalse(solution.tasks[task_1.name].scheduled)
        self.assertEqual(len(solution.tasks[task_1.name].assigned_resources), 0)

    def test_force_schedule_optional_tasks(self) -> None:
        """task_3 is not scheduled, because task_2 should not."""
        pb = ps.SchedulingProblem("ForceScheduleOptionalTasks", horizon=9)
        task_1 = ps.VariableDurationTask("task1", optional=True)
        task_2 = ps.FixedDurationTask("task2", duration=4, optional=True)
        task_3 = ps.FixedDurationTask("task3", duration=1, optional=True)

        cond = ps.ForceScheduleNOptionalTasks([task_1, task_2, task_3], 1)
        pb.add_constraint(cond)

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)

        results = [
            solution.tasks[task_1.name].scheduled,
            solution.tasks[task_2.name].scheduled,
            solution.tasks[task_3.name].scheduled,
        ]

        self.assertEqual(results.count(True), 1)

    def test_force_schedule_optional_tasks_2(self) -> None:
        """Just change the number of tasks to be scheduled."""
        pb = ps.SchedulingProblem("ForceScheduleOptionalTasks2", horizon=14)
        task_1 = ps.VariableDurationTask("task1", optional=True)
        task_2 = ps.FixedDurationTask("task2", duration=7, optional=True)
        task_3 = ps.FixedDurationTask("task3", duration=2, optional=True)

        cond = ps.ForceScheduleNOptionalTasks([task_1, task_2, task_3], 2)
        pb.add_constraint(cond)

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)

        results = [
            solution.tasks[task_1.name].scheduled,
            solution.tasks[task_2.name].scheduled,
            solution.tasks[task_3.name].scheduled,
        ]

        self.assertEqual(results.count(True), 2)

    def test_force_schedule_optional_tasks_3(self) -> None:
        """Check an error is raised if ever one of the task is not optional."""
        pb = ps.SchedulingProblem("ForceScheduleOptionalTasks3", horizon=14)
        task_1 = ps.VariableDurationTask("task1")  # this one is not optional
        task_2 = ps.FixedDurationTask("task2", duration=7, optional=True)

        with self.assertRaises(TypeError):
            cond = ps.ForceScheduleNOptionalTasks([task_1, task_2], 2)
            pb.add_constraint(cond)

    def test_get_scheduled_tasks(self) -> None:
        # task_1 cannot be scheduled, only tasks 2 and 3 can be
        pb = ps.SchedulingProblem("GetScheduledTasks", horizon=14)
        task_1 = ps.FixedDurationTask("task1", duration=15, optional=True)
        task_2 = ps.FixedDurationTask("task2", duration=7, optional=True)
        task_3 = ps.FixedDurationTask("task3", duration=2, optional=True)

        cond = ps.ForceScheduleNOptionalTasks([task_1, task_2, task_3], 2)
        pb.add_constraint(cond)

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()

        self.assertTrue(solution)

        scheduled_tasks_dictionary = solution.get_scheduled_tasks()

        self.assertEqual(len(scheduled_tasks_dictionary), 2)
        self.assertTrue("task2" in scheduled_tasks_dictionary)
        self.assertTrue("task3" in scheduled_tasks_dictionary)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import unittest

import processscheduler.jupyter_ui as jupui
from ipywidgets import VBox


class TestJupyerUI(unittest.TestCase):
    def test_load_jupyter_ui(self) -> None:
        self.assertTrue(isinstance(jupui.task_ui, VBox))
        self.assertTrue(isinstance(jupui.problem_ui, VBox))


if __name__ == "__main__":
    unittest.main()



"""The resources definition."""

# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, List, Optional, Tuple

from z3 import ArithRef, Bool, PbEq, PbGe, PbLe

from processscheduler.base import (
    _NamedUIDObject,
    is_positive_integer,
    is_strict_positive_integer,
)
from processscheduler.cost import _Cost, ConstantCostPerPeriod
import processscheduler.context as ps_context

#
# Utility functions
#
def _distribute_p_over_n(p, n):
    """Returns a list of integer p distributed over n values."""
    if p is None:
        return [p for i in range(n)]
    if isinstance(p, int):
        int_div = p // n
        to_return = [int_div + p % n]
        for _ in range(n - 1):
            to_return.append(int_div)
        return to_return
    elif isinstance(p, ConstantCostPerPeriod):
        int_div = p.value // n
        to_return = [ConstantCostPerPeriod(int_div + p.value % n)]
        for _ in range(n - 1):
            to_return.append(ConstantCostPerPeriod(int_div))
        return to_return


#
# Resources class definition
#
class _Resource(_NamedUIDObject):
    """base class for the representation of a resource"""

    def __init__(self, name: str) -> None:
        super().__init__(name)
        # for each resource, we define a dict that stores
        # all tasks and busy intervals of the resource.
        # busy intervals can be for example [(1,3), (5, 7)]
        self.busy_intervals = {}  # type: Dict[Task, Tuple[ArithRef, ArithRef]]

    def add_busy_interval(self, task, interval: Tuple[ArithRef, ArithRef]) -> None:
        """add an interval in which the resource is busy"""
        self.busy_intervals[task] = interval

    def get_busy_intervals(self) -> List[Tuple[ArithRef, ArithRef]]:
        """returns the list of all busy intervals"""
        return list(self.busy_intervals.values())


class Worker(_Resource):
    """A worker is an atomic resource that cannot be split into smaller parts.
    Typical workers are human beings, machines etc."""

    def __init__(
        self, name: str, productivity: Optional[int] = 1, cost: Optional[int] = None
    ) -> None:
        super().__init__(name)
        if not is_positive_integer(productivity):
            raise TypeError("productivity must be an integer >= 0")
        if cost is None:
            self.cost = None
        elif not isinstance(cost, _Cost):
            raise TypeError("cost must be a _Cost instance")
        self.productivity = productivity
        self.cost = cost

        # only worker are add to the main context, not SelectWorkers
        # add this resource to the current context
        if ps_context.main_context is None:
            raise AssertionError(
                "No context available. First create a SchedulingProblem"
            )
        ps_context.main_context.add_resource(self)


class SelectWorkers(_Resource):
    """Class representing the selection of n workers chosen among a list
    of possible workers"""

    def __init__(
        self,
        list_of_workers: List[_Resource],
        nb_workers_to_select: Optional[int] = 1,
        kind: Optional[str] = "exact",
    ):
        """create an instance of the SelectWorkers class."""
        super().__init__("")

        problem_function = {"min": PbGe, "max": PbLe, "exact": PbEq}

        if kind not in problem_function:
            raise ValueError("kind must be either 'exact', 'min' or 'max'")

        if not is_strict_positive_integer(nb_workers_to_select):
            raise TypeError("nb_workers must be an integer > 0")

        if nb_workers_to_select > len(list_of_workers):
            raise ValueError(
                "nb_workers must be <= the number of workers provided in list_of_workers."
            )

        # build the list of workers that will be the base of the selection
        # instances from this list mght either be Workers or CumulativeWorkers. If
        # this is a cumulative, then we add the list of all workers from the cumulative
        # into this list.
        self.list_of_workers = []
        for worker in list_of_workers:
            if isinstance(worker, CumulativeWorker):
                self.list_of_workers.extend(worker.cumulative_workers)
            else:
                self.list_of_workers.append(worker)

        # a dict that maps workers and selected boolean
        self.selection_dict = {}

        # create as many booleans as resources in the list
        for worker in self.list_of_workers:
            worker_is_selected = Bool("Selected_%s_%i" % (worker.name, self.uid))
            self.selection_dict[worker] = worker_is_selected

        # create the assertion : exactly n boolean flags are allowed to be True,
        # the others must be False
        # see https://github.com/Z3Prover/z3/issues/694
        # and https://stackoverflow.com/questions/43081929/k-out-of-n-constraint-in-z3py
        selection_list = list(self.selection_dict.values())
        self.selection_assertion = problem_function[kind](
            [(selected, True) for selected in selection_list], nb_workers_to_select
        )


class CumulativeWorker(_Resource):
    """A cumulative worker can process multiple tasks in parallel."""

    def __init__(
        self,
        name: str,
        size: int,
        productivity: Optional[int] = 1,
        cost: Optional[int] = None,
    ) -> None:
        super().__init__(name)

        if not (isinstance(size, int) and size >= 2):
            raise ValueError("CumulativeWorker 'size' attribute must be >=2.")

        if cost is None:
            self.cost = None

        elif not isinstance(cost, _Cost):
            raise TypeError("cost must be a _Cost instance")

        self.size = size
        # productivity and cost_per_period are distributed over
        # individual workers
        # for example, a productivty of 7 for a size of 3 will be distributed
        # as 3 on worker 1, 2 on worker 2 and 2 on worker 3. Same for cost_per_periods
        productivities = _distribute_p_over_n(productivity, size)

        costs_per_period = _distribute_p_over_n(cost, size)

        # we create as much elementary workers as the cumulative size
        self.cumulative_workers = [
            Worker(
                "%s_CumulativeWorker_%i" % (name, i + 1),
                productivity=productivities[i],
                cost=costs_per_period[i],
            )
            for i in range(size)
        ]

    def get_select_workers(self):
        """Each time the cumulative resource is assigned to a task, a SelectWorker
        instance is assigned to the task."""
        return SelectWorkers(
            self.cumulative_workers, nb_workers_to_select=1, kind="min"
        )



"""This module contains fundamental classes/functions that are used everywhere in the project."""

# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

from typing import List, Optional
import uuid
import warnings

from z3 import BoolRef, Bool, Implies, PbGe, PbEq, PbLe

#
# Utility functions
#
def is_strict_positive_integer(value: int) -> bool:
    """Return True if the parameter value is an integer > 0"""
    return isinstance(value, int) and value > 0


def is_positive_integer(value: int) -> bool:
    """Return True if the parameter value is an integer >= 0"""
    return isinstance(value, int) and value >= 0


#
# _NamedUIDObject, name and uid for hashing
#
class _NamedUIDObject:
    """The base object for most ProcessScheduler classes"""

    def __init__(self, name: str) -> None:
        """The base name for all ProcessScheduler objects.

        Provides an assertions list, a uniqueid.

        Args:
            name: the object name. It must be unique
        """
        # check name type
        if not isinstance(name, str):
            raise TypeError("name must be a str instance")
        # the object name
        self.name = name  # type: str

        # unique identifier
        self.uid = uuid.uuid4().int  # type: int

        # SMT assertions
        # start and end integer values must be positive
        self.assertions = []  # type: List[BoolRef]
        self.assertion_hashes = []

    def __hash__(self) -> int:
        return self.uid

    def __eq__(self, other) -> bool:
        return self.uid == other.uid

    def __repr__(self) -> str:
        """Print the object name, its uid and the assertions."""
        str_to_return = "%s(%s)\n%i assertion(s):\n" % (
            self.name,
            type(self),
            len(self.assertions),
        )
        assertions_str = "".join(["%s" % ass for ass in self.assertions])
        return str_to_return + assertions_str

    def add_assertion(self, z3_assertion: BoolRef) -> bool:
        """
        Add a z3 assertion to the list of assertions to be satisfied.

        Args:
            z3_assertion: the z3 assertion
        """
        # check if the assertion is in the list
        # workaround to avoid heavy hash computations
        assertion_hash = hash(z3_assertion)
        if assertion_hash in self.assertion_hashes:
            warnings.warn("assertion %s already added." % z3_assertion)
            return False
        self.assertions.append(z3_assertion)
        self.assertion_hashes.append(assertion_hash)
        return True

    def get_assertions(self) -> List[BoolRef]:
        """Return the assertions list"""
        return self.assertions


class _Constraint(_NamedUIDObject):
    """The base class for all constraints, including Task and Resource constraints."""

    def __init__(self, optional):
        super().__init__("")

        self.optional = optional
        # by default, this constraint has to be applied
        self.applied = True

    def set_assertions(self, list_of_z3_assertions: List[BoolRef]) -> None:
        """Take a list of constraint to satisfy. If the constraint is optional then
        the list of z3 assertions apply under the condition that the applied flag
        is set to True.
        """
        if self.optional:
            self.applied = Bool("constraint_%s_applied" % self.uid)
            self.add_assertion(Implies(self.applied, list_of_z3_assertions))
        else:
            self.applied = True
            self.add_assertion(list_of_z3_assertions)


class ForceApplyNOptionalConstraints(_Constraint):
    """Given a set of m different optional constraints, force the solver to apply
    at at least/at most/exactly n tasks, with 0 < n <= m. Work for both
    Task and/or Resource constraints."""

    def __init__(
        self,
        list_of_optional_constraints,
        nb_constraints_to_apply: Optional[int] = 1,
        kind: Optional[str] = "exact",
        optional: Optional[bool] = False,
    ) -> None:
        super().__init__(optional)

        problem_function = {"min": PbGe, "max": PbLe, "exact": PbEq}

        # first check that all tasks from the list_of_optional_tasks are
        # actually optional
        for constraint in list_of_optional_constraints:
            if not constraint.optional:
                raise TypeError(
                    "The constraint %s must explicitly be set as optional."
                    % constraint.name
                )

        # all scheduled variables to take into account
        applied_vars = []
        for constraint in list_of_optional_constraints:
            applied_vars.append(constraint.applied)

        asst = problem_function[kind](
            [(applied, True) for applied in applied_vars], nb_constraints_to_apply
        )
        self.set_assertions(asst)



# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

__VERSION__ = "0.7.1"

try:
    from z3 import BoolRef
except ModuleNotFoundError as z3_not_found:
    raise ImportError("z3 not found. It is a mandatory dependency") from z3_not_found

# Expose everything useful
from processscheduler.base import ForceApplyNOptionalConstraints
from processscheduler.first_order_logic import (
    not_,
    or_,
    and_,
    xor_,
    if_then_else,
    implies,
)
from processscheduler.objective import Indicator, MaximizeObjective, MinimizeObjective
from processscheduler.task import (
    ZeroDurationTask,
    FixedDurationTask,
    VariableDurationTask,
)
from processscheduler.task_constraint import *
from processscheduler.resource_constraint import (
    SameWorkers,
    DistinctWorkers,
    ResourceUnavailable,
    WorkLoad,
)
from processscheduler.resource import Worker, CumulativeWorker, SelectWorkers
from processscheduler.cost import ConstantCostPerPeriod, PolynomialCostFunction
from processscheduler.problem import SchedulingProblem
from processscheduler.solver import SchedulingSolver
from processscheduler.context import main_context, SchedulingContext, clear_main_context



# ProcessScheduler benchmark
import sys
import time
from datetime import datetime
import subprocess
import os
import platform
import psutil
import uuid

import matplotlib.pyplot as plt
import processscheduler as ps
import z3

#
# Display machine identification
#
def get_size(byt, suffix="B"):
    """
    Scale bytes to its proper format
    e.g:
        1253656 => '1.20MB'
        1253656678 => '1.17GB'
    """
    # Code from https://www.thepythoncode.com/article/get-hardware-system-information-python
    factor = 1024
    for unit in ["", "K", "M", "G", "T", "P"]:
        if byt < factor:
            return f"{byt:.2f}{unit}{suffix}"
        byt /= factor


bench_id = uuid.uuid4().hex
bench_date = datetime.now()
print("#### Benchmark information header ####")
print("Date:", bench_date)
print("Id:", bench_id)
print("Software:")
print("\tPython version:", platform.python_version())
print("\tProcessScheduler version:", ps.__VERSION__)
commit_short_hash = subprocess.check_output(
    ["git", "rev-parse", "--short", "HEAD"]
).strip()
print("\tz3 version:", z3.Z3_get_full_version())

print("\tProcessScheduler commit number:", commit_short_hash.decode("utf-8"))
os_info = os.uname()
print("OS:")
print("\tOS:", os_info.sysname)
print("\tOS Release:", os_info.release)
print("\tOS Version:", os_info.version)
print("Hardware:")
print("\tMachine:", os_info.machine)
print("\tPhysical cores:", psutil.cpu_count(logical=False))
print("\tTotal cores:", psutil.cpu_count(logical=True))
# CPU frequencies
cpufreq = psutil.cpu_freq()
print(f"\tMax Frequency: {cpufreq.max:.2f}Mhz")
print(f"\tMin Frequency: {cpufreq.min:.2f}Mhz")
# get the memory details
svmem = psutil.virtual_memory()
print(f"\tTotal memory: {get_size(svmem.total)}")

model_creation_times = []
computation_times = []

n = 30  # max number of dev teams

N = list(range(4, n, 2))  # from 4 to N, step 2
test_init_time = time.perf_counter()
for num_dev_teams in N:
    print("-> Num dev teams:", num_dev_teams)
    # Teams and Resources
    num_resource_a = 2
    num_resource_b = 2

    init_time = time.perf_counter()
    # Resources
    digital_transformation = ps.SchedulingProblem("DigitalTransformation")
    r_a = ps.Worker("A")

    # Dev Team Tasks
    # For each dev_team pick one resource a and one resource b.
    ts_team_migration = [
        ps.FixedDurationTask("DevTeam_%i" % (i + 1), duration=1, priority=10)
        for i in range(num_dev_teams)
    ]
    for t_team_migration in ts_team_migration:
        t_team_migration.add_required_resource(r_a)

    # solve
    # digital_transformation.add_objective_priorities()
    digital_transformation.add_objective_makespan()

    top1 = time.perf_counter()
    model_creation_time = top1 - init_time

    # print("Teams: ", num_dev_teams)
    # print("Creating ")
    mt = 5
    solver = ps.SchedulingSolver(digital_transformation, random_seed=True, max_time=mt)
    # print("Done ok.")
    # print("Solve.")
    top2 = time.perf_counter()
    solution = solver.solve()

    computing_time = time.perf_counter() - top2
    if computing_time > mt:
        computing_time = computing_time - mt

    model_creation_times.append(model_creation_time)
    computation_times.append(computing_time)

    solver.print_statistics()
test_final_time = time.perf_counter()
print("TOTAL BENCH TIME:", test_final_time - test_init_time)
plt.title("Benchmark SelectWorkers %s:%s" % (bench_date, bench_id[8:]))
plt.plot(model_creation_times, label="Model creation")
plt.plot(computation_times, label="Computing time")
plt.legend()
plt.grid(True)
plt.show()



"""The module that defines the task related classes."""

# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

from typing import List, Optional

from z3 import Bool, BoolRef, Int, And, If

from processscheduler.base import (
    _NamedUIDObject,
    is_strict_positive_integer,
    is_positive_integer,
)
from processscheduler.resource import _Resource, Worker, CumulativeWorker, SelectWorkers

import processscheduler.context as ps_context


class Task(_NamedUIDObject):
    """a Task object"""

    def __init__(self, name: str, optional: bool) -> None:
        super().__init__(name)
        self.work_amount = 0
        self.priority = 1  # by default

        # required resources to perform the task
        self.required_resources = []  # type: List[_Resource]

        # z3 Int variables
        self.start = Int("%s_start" % name)
        self.end = Int("%s_end" % name)
        self.duration = Int("%s_duration" % name)

        # by default, this task has to be scheduled
        self.optional = optional
        self.scheduled = True

        # add this task to the current context
        if ps_context.main_context is None:
            raise AssertionError(
                "No context available. First create a SchedulingProblem"
            )
        ps_context.main_context.add_task(self)

        # the counter used for negative integers
        self._current_negative_integer = 0

    def _get_unique_negative_integer(self):
        """Returns a new negative integer each time this method is called."""
        # decrement the current counter
        self._current_negative_integer -= 1
        return self._current_negative_integer

    def add_required_resource(self, resource: _Resource, dynamic=False) -> None:
        """
        Add a required resource to the current task.

        Args:
            resource: any of one of the _Resource derivatives class (Worker, SelectWorkers etc.)
        If dynamic flag (False by default) is set to True, then the resource is dynamic
        and can join the task any time between its start and end times.
        """
        if not isinstance(resource, _Resource):
            raise TypeError("you must pass a Resource instance")

        if resource in self.required_resources:
            raise ValueError(
                "resource %s already defined as a required resource for task %s"
                % (resource.name, self.name)
            )

        if isinstance(resource, CumulativeWorker):
            # in the case for a CumulativeWorker, select at least one worker
            resource = resource.get_select_workers()

        if isinstance(resource, SelectWorkers):
            # loop over each resource
            for worker in resource.list_of_workers:
                resource_maybe_busy_start = Int(
                    "%s_maybe_busy_%s_start" % (worker.name, self.name)
                )
                resource_maybe_busy_end = Int(
                    "%s_maybe_busy_%s_end" % (worker.name, self.name)
                )
                # create the busy interval for the resource
                worker.add_busy_interval(
                    self, (resource_maybe_busy_start, resource_maybe_busy_end)
                )
                # add assertions. If worker is selected then sync the resource with the task
                selected_variable = resource.selection_dict[worker]
                schedule_as_usual = And(
                    resource_maybe_busy_start == self.start,
                    resource_maybe_busy_end == self.end,
                )
                # in the case the worker is selected
                # move the busy interval to a single point in time, in the
                # past. This way, it does not conflict with tasks to be
                # actuallt scheduled.
                # This single point in time results in a zero duration time: related
                # task will not be considered when cimputing resource utilization or cost.
                single_point_in_past = self._get_unique_negative_integer()
                move_to_past = And(
                    resource_maybe_busy_start == single_point_in_past,  # to past
                    resource_maybe_busy_end == single_point_in_past,
                )
                # define the assertion ...
                assertion = If(selected_variable, schedule_as_usual, move_to_past)
                # ... and store it into the task assertions list
                self.add_assertion(assertion)
                # finally, add each worker to the "required" resource list
                self.required_resources.append(worker)
            # also, don't forget to add the AlternativeWorker assertion
            self.add_assertion(resource.selection_assertion)
        elif isinstance(resource, Worker):
            resource_busy_start = Int("%s_busy_%s_start" % (resource.name, self.name))
            resource_busy_end = Int("%s_busy_%s_end" % (resource.name, self.name))
            # create the busy interval for the resource
            resource.add_busy_interval(self, (resource_busy_start, resource_busy_end))
            # set the busy resource to keep synced with the task
            if dynamic:
                self.add_assertion(resource_busy_end <= self.end)
                self.add_assertion(resource_busy_start >= self.start)
            else:
                # self.add_assertion(resource_busy_start + self.duration == resource_busy_end)
                self.add_assertion(resource_busy_end == self.end)
                self.add_assertion(resource_busy_start == self.start)
            # finally, store this resource into the resource list
            self.required_resources.append(resource)

    def add_required_resources(
        self, list_of_resources: List[_Resource], dynamic=False
    ) -> None:
        """
        Add a set of required resources to the current task.

        This method calls the add_required_resource method for each resource of the list.
        As a consequence, be aware this is not an atomic transaction.

        Args:
            list_of_resources: the list of resources to add.
        """
        for resource in list_of_resources:
            self.add_required_resource(resource, dynamic)

    def set_assertions(self, list_of_z3_assertions: List[BoolRef]) -> None:
        """Take a list of constraint to satisfy. Create two cases: if the task is scheduled,
        nothing is done, if the case is optional, scheduling the task to the past"""
        if self.optional:  # in this case the previous assertions maybe skipped
            self.scheduled = Bool("%s_scheduled" % self.name)
            not_scheduled_assertion = And(
                self.start == -1, self.end == -1, self.duration == 0  # to past
            )
            self.add_assertion(
                If(self.scheduled, And(list_of_z3_assertions), not_scheduled_assertion)
            )
        else:
            self.scheduled = True
            self.add_assertion(And(list_of_z3_assertions))


class ZeroDurationTask(Task):
    """Task with zero duration, an instant in the schedule.

    The task end and start are constrained to be equal.

    Args:
        name: the task name. It must be unique
    """

    def __init__(self, name: str, optional: Optional[bool] = False) -> None:
        super().__init__(name, optional)
        # add an assertion: end = start because the duration is zero
        assertions = [self.start == self.end, self.duration == 0]

        self.set_assertions(assertions)


class FixedDurationTask(Task):
    """Task with constant duration.

    Args:
        name: the task name. It must be unique
        duration: the task duration as a number of periods
        work_amount: represent the quantity of work this task must produce
        priority: the task priority. The greater the priority, the sooner it will be scheduled
        optional: True if task schedule is optional, False otherwise (default)
    """

    def __init__(
        self,
        name: str,
        duration: int,
        work_amount: Optional[int] = 0,
        priority: Optional[int] = 1,
        optional: Optional[bool] = False,
    ) -> None:
        super().__init__(name, optional)
        if not is_strict_positive_integer(duration):
            raise TypeError("duration must be a strict positive integer")
        if not is_positive_integer(work_amount):
            raise TypeError("work_amount me be a positive integer")

        self.work_amount = work_amount
        self.priority = priority

        assertions = [
            self.start + self.duration == self.end,
            self.duration == duration,
            self.start >= 0,
        ]

        self.set_assertions(assertions)


class UnavailabilityTask(FixedDurationTask):
    """A task that tells that a resource is unavailable during this period. This
    task is not publicly exposed, it is used by the resource constraint
    ResourceUnavailability.
    """

    def __init__(self, name: str, duration: int) -> None:
        super().__init__(name, duration, work_amount=0, priority=0)


class VariableDurationTask(Task):
    """Task with a priori unknown duration. its duration is computed by the solver"""

    def __init__(
        self,
        name: str,
        min_duration: Optional[int] = 0,
        max_duration: Optional[int] = None,
        work_amount: Optional[int] = 0,
        priority: Optional[int] = 1,
        optional: Optional[bool] = False,
    ):
        super().__init__(name, optional)

        if is_positive_integer(max_duration):
            self.add_assertion(self.duration <= max_duration)
        elif max_duration is not None:
            raise TypeError(
                "length_as_most should either be a positive integer or None"
            )

        if not is_positive_integer(min_duration):
            raise TypeError("min_duration must be a positive integer")

        if not is_positive_integer(work_amount):
            raise TypeError("work_amount me be a positive integer")

        self.min_duration = min_duration
        self.max_duration = max_duration
        self.work_amount = work_amount
        self.priority = priority

        assertions = [
            self.start + self.duration == self.end,
            self.start >= 0,
            self.duration >= min_duration,
        ]

        if max_duration is not None:
            assertions.append(self.duration <= max_duration)

        if min_duration is not None:
            assertions.append(self.duration >= min_duration)

        self.set_assertions(assertions)



# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import unittest

import processscheduler as ps
from processscheduler.resource import _distribute_p_over_n


class TestCumulative(unittest.TestCase):
    def test_distribute_p_over_n(self):
        res1 = _distribute_p_over_n(7, 3)
        self.assertEqual(res1, [3, 2, 2])
        res2 = _distribute_p_over_n(6, 2)
        self.assertEqual(res2, [3, 3])
        res3 = _distribute_p_over_n(9, 3)
        self.assertEqual(res3, [3, 3, 3])
        res4 = _distribute_p_over_n(1, 5)
        self.assertEqual(res4, [1, 0, 0, 0, 0])
        res4 = _distribute_p_over_n(29, 3)
        self.assertEqual(res4, [11, 9, 9])

    def test_create_cumulative(self):
        """take the single task/single resource and display output"""
        ps.SchedulingProblem("CreateCumulative", horizon=10)
        cumulative_worker = ps.CumulativeWorker("MachineA", size=4)
        self.assertEqual(len(cumulative_worker.cumulative_workers), 4)

    def test_create_cumulative_wrong_type(self):
        """take the single task/single resource and display output"""
        ps.SchedulingProblem("CreateCumulativeWrongType", horizon=10)
        with self.assertRaises(ValueError):
            ps.CumulativeWorker("MachineA", size=1)
        with self.assertRaises(ValueError):
            ps.CumulativeWorker("MachineA", size=2.5)

    def test_cumulative_1(self):
        pb_bs = ps.SchedulingProblem("Cumulative1", 3)
        # tasks
        t1 = ps.FixedDurationTask("T1", duration=2)
        t2 = ps.FixedDurationTask("T2", duration=2)
        t3 = ps.FixedDurationTask("T3", duration=2)

        # workers
        r1 = ps.CumulativeWorker("Machine1", size=3)
        # resource assignment
        t1.add_required_resource(r1)
        t2.add_required_resource(r1)
        t3.add_required_resource(r1)

        # constraints
        pb_bs.add_constraint(ps.TaskStartAt(t2, 1))

        # plot solution
        solver = ps.SchedulingSolver(pb_bs)
        solution = solver.solve()
        self.assertTrue(solution)

    def test_cumulative_2(self):
        pb_bs = ps.SchedulingProblem("Cumulative2", 3)
        # tasks
        t1 = ps.FixedDurationTask("T1", duration=2)
        t2 = ps.FixedDurationTask("T2", duration=2)

        # workers
        r1 = ps.CumulativeWorker("Machine1", size=2)
        # resource assignment
        t1.add_required_resource(r1)
        t2.add_required_resource(r1)

        # constraints
        pb_bs.add_constraint(ps.TaskStartAt(t2, 1))

        # plot solution
        solver = ps.SchedulingSolver(pb_bs)
        solution = solver.solve()
        self.assertTrue(solution)

    def test_optional_cumulative(self):
        """Same as above, but with an optional taskand an horizon of 2.
        t2 should not be scheduled."""
        pb_bs = ps.SchedulingProblem("OptionalCumulative", 2)
        # tasks
        t1 = ps.FixedDurationTask("T1", duration=2)
        t2 = ps.FixedDurationTask("T2", duration=2, optional=True)
        t3 = ps.FixedDurationTask("T3", duration=2)

        # workers
        r1 = ps.CumulativeWorker("Machine1", size=2)
        # resource assignment
        t1.add_required_resource(r1)
        t2.add_required_resource(r1)

        # constraints
        pb_bs.add_constraint(ps.TaskStartAt(t2, 1))

        # plot solution
        solver = ps.SchedulingSolver(pb_bs)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertTrue(solution.tasks[t1.name].scheduled)
        self.assertTrue(solution.tasks[t3.name].scheduled)
        self.assertFalse(solution.tasks[t2.name].scheduled)

    def test_cumulative_select_worker_1(self):
        pb_bs = ps.SchedulingProblem("CumulativeSelectWorker", 2)
        # tasks
        t1 = ps.FixedDurationTask("T1", duration=2)
        t2 = ps.FixedDurationTask("T2", duration=2)

        # workers
        r1 = ps.CumulativeWorker("Machine1", size=2)
        r2 = ps.CumulativeWorker("Machine2", size=2)
        # resource assignment
        t1.add_required_resource(ps.SelectWorkers([r1, r2], 1))
        t2.add_required_resource(ps.SelectWorkers([r1, r2], 1))

        # plot solution
        solver = ps.SchedulingSolver(pb_bs)
        solution = solver.solve()
        self.assertTrue(solution)

    def test_cumulative_hosp(self):
        n = 16
        capa = 4
        pb_bs = ps.SchedulingProblem("Hospital", horizon=int(n / capa))
        # workers
        r1 = ps.CumulativeWorker("Room", size=capa)

        for i in range(n):
            t = ps.FixedDurationTask("T%i" % (i + 1), duration=1)
            t.add_required_resource(r1)

        solver = ps.SchedulingSolver(pb_bs)
        solution = solver.solve()
        self.assertTrue(solution)

    def test_cumulative_productivity(self):
        """Horizon should be 4, 100/29=3.44."""
        problem = ps.SchedulingProblem("CumulativeProductivity")
        t_1 = ps.VariableDurationTask("t1", work_amount=100)

        worker_1 = ps.CumulativeWorker("CumulWorker", size=3, productivity=29)
        t_1.add_required_resource(worker_1)

        problem.add_objective_makespan()
        solution = ps.SchedulingSolver(problem).solve()
        self.assertTrue(solution)
        self.assertEqual(solution.horizon, 4)


if __name__ == "__main__":
    unittest.main()



"""Solution definition and Gantt export."""

# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

from datetime import time, timedelta, datetime
import json
from pathlib import Path
import random

from typing import Optional, Tuple


class SolutionJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime) or isinstance(obj, (time, timedelta)):
            return "%s" % obj
        return obj.__dict__


class TaskSolution:
    """Class to represent the solution for a scheduled Task."""

    def __init__(self, name: str):
        self.name = name
        self.type = ""  # the name of the task type
        self.start = 0  # an integer
        self.end = 0  # an integer
        self.duration = 0  # an integer

        self.start_time = ""  # a string
        self.end_time = ""  # a string
        self.duration_time = ""  # a string

        self.optional = False
        self.scheduled = False
        # the name of assigned resources
        self.assigned_resources = []


class ResourceSolution:
    """Class to represent the solution for the resource assignments."""

    def __init__(self, name: str):
        self.name = name
        self.type = ""  # the name of the task type
        # an assignment is a list of tuples : [(Task_name, start, end), (task_2name, start2, end2) etc.]
        self.assignments = []


class SchedulingSolution:
    """A class that represent the solution of a scheduling problem. Can be rendered
    to a matplotlib Gantt chart, or exported to json
    """

    def __init__(self, problem):
        """problem: a scheduling problem."""
        self.problem_name = problem.name
        self.problem = problem
        self.horizon = 0
        self.tasks = {}  # the dict of tasks
        self.resources = {}  # the dict of all resources
        self.indicators = {}  # the dict of inicators values

    def __repr__(self):
        return self.to_json_string()

    def get_all_tasks_but_unavailable(self):
        """Return all tasks except those of the type UnavailabilityTask
        used to represent a ResourceUnavailable constraint."""
        tasks_to_return = {}
        for task in self.tasks:
            if "NotAvailable" not in task:
                tasks_to_return[task] = self.tasks[task]
        return tasks_to_return

    def get_scheduled_tasks(self):
        """Return scheduled tasks."""
        tasks_not_unavailable = self.get_all_tasks_but_unavailable()
        tasks_to_return = {}
        for task in tasks_not_unavailable:
            if tasks_not_unavailable[task].scheduled:
                tasks_to_return[task] = tasks_not_unavailable[task]
        return tasks_to_return

    def to_json_string(self) -> str:
        """Export the solution to a json string."""
        d = {}
        # problem properties
        problem_properties = {}
        d["horizon"] = self.horizon
        # time data
        problem_properties["problem_timedelta"] = self.problem.delta_time
        if self.problem.delta_time is not None:
            if self.problem.start_time is not None:
                problem_properties["problem_start_time"] = self.problem.start_time
                problem_properties["problem_end_time"] = (
                    self.problem.start_time + self.horizon * self.problem.delta_time
                )
            else:
                problem_properties["problem_start_time"] = time(0)
                problem_properties["problem_end_time"] = (
                    self.horizon * self.problem.delta_time
                )
        else:
            problem_properties["problem_start_time"] = None
            problem_properties["problem_end_time"] = None
        d["problem_properties"] = problem_properties

        d["tasks"] = self.tasks
        d["resources"] = self.resources
        d["indicators"] = self.indicators
        return json.dumps(d, indent=4, sort_keys=True, cls=SolutionJSONEncoder)

    def add_indicator_solution(self, indicator_name: str, indicator_value: int) -> None:
        """Add indicator solution."""
        self.indicators[indicator_name] = indicator_value

    def add_task_solution(self, task_solution: TaskSolution) -> None:
        """Add task solution."""
        self.tasks[task_solution.name] = task_solution

    def add_resource_solution(self, resource_solution: ResourceSolution) -> None:
        """Add resource solution."""
        self.resources[resource_solution.name] = resource_solution

    def render_gantt_plotly(
        self,
        fig_size: Optional[Tuple[int, int]] = None,
        show_plot: Optional[bool] = True,
        show_indicators: Optional[bool] = True,
        render_mode: Optional[str] = "Resource",
        sort: Optional[str] = None,
        fig_filename: Optional[str] = None,
        html_filename: Optional[str] = None,
    ) -> None:
        """Use plotly.create_gantt method, see
        https://plotly.github.io/plotly.py-docs/generated/plotly.figure_factory.create_gantt.html
        """
        try:
            from plotly.figure_factory import create_gantt
        except ImportError:
            raise ModuleNotFoundError("plotly is not installed.")

        if render_mode not in ["Task", "Resource"]:
            raise ValueError("data_type must be either Task or Resource")

        # tasks to render
        if render_mode == "Task":
            tasks_to_render = self.get_all_tasks_but_unavailable()
        else:
            tasks_to_render = self.tasks

        df = []
        for task_name in tasks_to_render:
            task_solution = self.tasks[task_name]
            if task_solution.assigned_resources:
                resource_text = ",".join(task_solution.assigned_resources)
            else:
                resource_text = r"($\emptyset$)"
            df.append(
                dict(
                    Task=task_name,
                    Start=task_solution.start_time,
                    Finish=task_solution.end_time,
                    Resource=resource_text,
                )
            )

        gantt_title = "%s Gantt chart" % self.problem.name
        # add indicators value to title
        if self.indicators and show_indicators:
            for indicator_name in self.indicators:
                gantt_title += " - %s: %i" % (
                    indicator_name,
                    self.indicators[indicator_name],
                )

        r = lambda: random.randint(0, 255)
        colors = []
        for _ in range(len(df)):
            colors.append("#%02X%02X%02X" % (r(), r(), r()))

        if sort is not None:
            if sort in ["Task", "Resource"]:
                df = sorted(df, key=lambda i: i[sort], reverse=False)
            elif sort in ["Start", "Finish"]:
                df = sorted(df, key=lambda i: i[sort], reverse=True)
            else:
                raise ValueError(
                    'sort must be either "Task", "Resource", "Start", or "Finish"'
                )

        if fig_size is None:
            fig = create_gantt(
                df,
                colors=colors,
                index_col=render_mode,
                show_colorbar=True,
                showgrid_x=True,
                showgrid_y=True,
                show_hover_fill=True,
                title=gantt_title,
                bar_width=0.5,
            )
        else:
            fig = create_gantt(
                df,
                colors=colors,
                index_col=render_mode,
                show_colorbar=True,
                showgrid_x=True,
                showgrid_y=True,
                show_hover_fill=True,
                title=gantt_title,
                bar_width=0.5,
                width=fig_size[0],
                height=fig_size[1],
            )

        if fig_filename is not None:
            fig.write_image(fig_filename)

        if html_filename is not None:
            file = Path(html_filename)
            file.write_text(fig.to_html(include_plotlyjs="cdn"))

        if show_plot:
            fig.show()

    def render_gantt_matplotlib(
        self,
        fig_size: Optional[Tuple[int, int]] = (9, 6),
        show_plot: Optional[bool] = True,
        show_indicators: Optional[bool] = True,
        render_mode: Optional[str] = "Resource",
        fig_filename: Optional[str] = None,
    ) -> None:
        """generate a gantt diagram using matplotlib.
        Inspired by
        https://www.geeksforgeeks.org/python-basic-gantt-chart-using-matplotlib/
        """
        try:
            import matplotlib.pyplot as plt
            from matplotlib.colors import LinearSegmentedColormap
        except ImportError:
            raise ModuleNotFoundError("matplotlib is not installed.")

        if not self.resources:
            render_mode = "Task"

        if render_mode not in ["Resource", "Task"]:
            raise ValueError("render_mode must either be 'Resource' or 'Task'")

        # tasks to render
        if render_mode == "Task":
            tasks_to_render = self.get_all_tasks_but_unavailable()
        else:
            tasks_to_render = self.tasks

        # render mode is Resource by default, can be set to 'Task'
        if render_mode == "Resource":
            plot_title = "Resources schedule - %s" % self.problem_name
            plot_ylabel = "Resources"
            plot_ticklabels = list(self.resources.keys())
            nbr_y_values = len(self.resources)
        elif render_mode == "Task":
            plot_title = "Task schedule - %s" % self.problem_name
            plot_ylabel = "Tasks"
            plot_ticklabels = list(tasks_to_render.keys())
            nbr_y_values = len(tasks_to_render)

        gantt = plt.subplots(1, 1, figsize=fig_size)[1]
        gantt.set_title(plot_title)

        # x axis, use real date and times if possible
        if self.problem.delta_time is not None:
            if self.problem.start_time is not None:
                # get all days
                times = [
                    self.problem.start_time + i * self.problem.delta_time
                    for i in range(self.horizon + 1)
                ]
                times_str = []
                for t in times:
                    times_str.append(t.strftime("%H:%M"))
            else:
                times_str = [
                    "%s" % (i * self.problem.delta_time)
                    for i in range(self.horizon + 1)
                ]
            gantt.set_xlim(0, self.horizon)
            plt.xticks(range(self.horizon + 1), times_str, rotation=60)
            plt.subplots_adjust(bottom=0.15)
            gantt.set_xlabel("Time", fontsize=12)
        else:
            # otherwise use integers
            gantt.set_xlim(0, self.horizon)
            gantt.set_xticks(range(self.horizon + 1))
            # Setting label
            gantt.set_xlabel("Time (%i periods)" % self.horizon, fontsize=12)

        gantt.set_ylabel(plot_ylabel, fontsize=12)

        # colormap definition
        cmap = LinearSegmentedColormap.from_list(
            "custom blue", ["#bbccdd", "#ee3300"], N=len(self.tasks)
        )  # nbr of colors
        # defined a mapping between the tasks and the colors, so that
        # each task has the same color on both graphs
        task_colors = {}
        for i, task_name in enumerate(tasks_to_render):
            task_colors[task_name] = cmap(i)
        # the task color is defined from the task name, this way the task has
        # already the same color, even if it is defined after
        gantt.set_ylim(0, 2 * nbr_y_values)
        gantt.set_yticks(range(1, 2 * nbr_y_values, 2))
        gantt.set_yticklabels(plot_ticklabels)
        # in Resources mode, create one line per resource on the y axis
        gantt.grid(axis="x", linestyle="dashed")

        def draw_broken_barh_with_text(start, length, bar_color, text, hatch=None):
            # first compute the bar dimension
            if length == 0:  # zero duration tasks, to be visible
                bar_dimension = (start - 0.05, 0.1)
            else:
                bar_dimension = (start, length)
            gantt.broken_barh(
                [bar_dimension],
                (i * 2, 2),
                edgecolor="black",
                linewidth=2,
                facecolors=bar_color,
                hatch=hatch,
                alpha=0.5,
            )
            gantt.text(
                x=start + length / 2,
                y=i * 2 + 1,
                s=text,
                ha="center",
                va="center",
                color="black",
            )

        # in Tasks mode, create one line per task on the y axis
        if render_mode == "Task":
            for i, task_name in enumerate(tasks_to_render):
                # build the bar text string
                task_solution = self.tasks[task_name]
                if task_solution.assigned_resources:
                    text = ",".join(task_solution.assigned_resources)
                else:
                    text = r"($\emptyset$)"
                draw_broken_barh_with_text(
                    task_solution.start,
                    task_solution.duration,
                    task_colors[task_name],
                    text,
                )
        elif render_mode == "Resource":
            for i, resource_name in enumerate(self.resources):
                ress = self.resources[resource_name]
                # each interval from the busy_intervals list is rendered as a bar
                for task_name, start, end in ress.assignments:
                    # unavailabilities are rendered with a grey dashed bar
                    if "NotAvailable" in task_name:
                        hatch = "//"
                        bar_color = "white"
                        text_to_display = ""
                    else:
                        hatch = None
                        bar_color = task_colors[task_name]
                        text_to_display = task_name
                    draw_broken_barh_with_text(
                        start, end - start, bar_color, text_to_display, hatch
                    )
        # display indicator values in the legend area
        if self.indicators and show_indicators:
            for indicator_name in self.indicators:
                gantt.plot(
                    [],
                    [],
                    " ",
                    label="%s: %i" % (indicator_name, self.indicators[indicator_name]),
                )
            gantt.legend(title="Indicators", title_fontsize="large", framealpha=0.5)

        if fig_filename is not None:
            plt.savefig(fig_filename)

        if show_plot:
            plt.show()



"""The context related module."""

# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, List, Union
import warnings

from z3 import BoolRef, ArithRef

from processscheduler.task_constraint import _Constraint


class SchedulingContext:
    """A class that contains all related data resource, tasks etc.

    The methods defined in this class ensures
    """

    def __init__(self):
        # set and clear variables
        self.clear()

    def clear(self):
        # the list of tasks to be scheduled in this scenario
        self.tasks = []  # type: Dict[str, Task]
        # the list of resources available in this scenario
        self.resources = []  # type: Dict[str, _Resource]
        # the constraints are defined in the scenario
        self.constraints = []  # type: List[BoolRef]
        # list of define indicators
        self.indicators = []  # type: List[Indicator]
        self.objectives = []  # type: List[Union[Indicator, ArithRef]]

    def add_indicator(self, indicator) -> bool:
        """add an indicatr to the problem"""
        if indicator not in self.indicators:
            self.indicators.append(indicator)
        else:
            warnings.warn("indicator %s already part of the problem" % indicator)
            return False
        return True

    def add_task(self, task) -> None:
        """add a single task to the problem. There must not be two tasks with the same name"""
        all_task_names = [t.name for t in self.tasks]
        if task.name in all_task_names:
            raise ValueError("a task with the name %s already exists." % task.name)
        self.tasks.append(task)

    def add_resource(self, resource) -> None:
        """add a single resource to the problem"""
        all_resource_names = [t.name for t in self.resources]
        if resource.name in all_resource_names:
            raise ValueError(
                "a resource with the name %s already exists." % resource.name
            )
        self.resources.append(resource)

    def add_constraint(self, constraint) -> None:
        """add a constraint to the problem"""
        if isinstance(constraint, _Constraint):
            self.constraints.append(constraint.get_assertions())
        elif isinstance(constraint, BoolRef):
            self.constraints.append(constraint)
        else:
            raise TypeError(
                "You must provide either a _Constraint or BoolRef instance."
            )

    def add_objective(self, objective) -> None:
        self.objectives.append(objective)


# Define a global context
# None by default
# the scheduling problem will set
# this variable
main_context = None


def clear_main_context() -> None:
    main_context.clear()



"""First order logic operators, implies, if/then/else."""

# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

from typing import Union, List

from z3 import And, Xor, Or, Not, If, Implies, BoolRef

from processscheduler.base import _NamedUIDObject

#
# Utility functions
#
def _get_assertions(constraint: Union[BoolRef, _NamedUIDObject]) -> BoolRef:
    """Take a BoolRef or any _NamedUID and returns the assertions for this object."""
    if isinstance(constraint, BoolRef):
        assertion = constraint
    elif isinstance(constraint, _NamedUIDObject):
        assertion = constraint.get_assertions()
    else:
        raise TypeError(
            "constraint must either be a _NamedUIDObject or BoolRef instance"
        )
    return assertion


def _constraints_to_list_of_assertions(list_of_constraints) -> List[BoolRef]:
    """Convert a list of constraints or assertions to a list of assertions."""
    list_of_boolrefs_to_return = []
    for constraint in list_of_constraints:
        assertions = _get_assertions(constraint)
        if isinstance(assertions, list):
            list_of_boolrefs_to_return.extend(assertions)
        elif isinstance(assertions, BoolRef):
            list_of_boolrefs_to_return.append(assertions)
    return list_of_boolrefs_to_return


#
# Nested boolean operators for _NamedUIDObject objects
# or BoolRef
#
def not_(constraint: Union[BoolRef, _NamedUIDObject]) -> BoolRef:
    """Boolean negation of the constraint."""
    return Not(And(_get_assertions(constraint)))


def or_(list_of_constraints: List[Union[BoolRef, _NamedUIDObject]]) -> BoolRef:
    """Boolean 'or' between a list of assertions or constraints.

    At least one assertion in the list must be satisfied.
    """
    return Or(_constraints_to_list_of_assertions(list_of_constraints))


def and_(list_of_constraints: List[Union[BoolRef, _NamedUIDObject]]) -> BoolRef:
    """Boolean 'and' between a list of assertions or constraints.

    All assertions must be satisfied.
    """
    return And(_constraints_to_list_of_assertions(list_of_constraints))


def xor_(list_of_constraints: List[Union[BoolRef, _NamedUIDObject]]) -> BoolRef:
    """Boolean 'xor' between two assertions or constraints.

    One assertion must be satisfied, the other is not satisfied. The list of constraint
    must have exactly 2 elements.
    """
    if len(list_of_constraints) != 2:
        raise TypeError(
            "You list size must be 2. Be sure you have 2 constraints in the list."
        )

    constraint_1 = list_of_constraints[0]
    constraint_2 = list_of_constraints[1]

    return Xor(And(_get_assertions(constraint_1)), And(_get_assertions(constraint_2)))


#
# Logical consequence
#
def implies(
    condition: Union[BoolRef, _NamedUIDObject],
    consequence_list_of_constraints: List[Union[BoolRef, _NamedUIDObject]],
) -> BoolRef:
    """Return an implie instance

    Args:
        condition: a constraint or a boolref
        consequence_list_of_constraints: a list of all implications if condition is True
    """
    return Implies(
        And(_get_assertions(condition)),
        And(_constraints_to_list_of_assertions(consequence_list_of_constraints)),
    )


#
# If/then/else
#
def if_then_else(
    condition: Union[BoolRef, _NamedUIDObject],
    then_list_of_constraints: List[Union[BoolRef, _NamedUIDObject]],
    else_list_of_constraints: List[Union[BoolRef, _NamedUIDObject]],
) -> BoolRef:
    """If/Then/Else statement.

    Args:
        condition: a constraint or a boolref
        then_list_of_constraints: a list of all implications if condition is True
        else_list_of_constraints: a list of all implications if condition is False
    """
    return If(
        And(_get_assertions(condition)),
        And(_constraints_to_list_of_assertions(then_list_of_constraints)),
        And(_constraints_to_list_of_assertions(else_list_of_constraints)),
    )



# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import unittest

import processscheduler as ps
import processscheduler.context as ps_context


def new_problem_or_clear() -> None:
    """clear the current context. If no context is defined,
    create a SchedulingProject object"""
    if ps_context.main_context is None:
        ps.SchedulingProblem("NewProblem")
    else:
        ps_context.main_context.clear()


class TestTask(unittest.TestCase):
    def test_clear_context(self) -> None:
        ps_context.main_context = None
        new_problem_or_clear()
        self.assertIsInstance(ps_context.main_context, ps.SchedulingContext)

    def test_create_task_without_problem(self) -> None:
        ps_context.main_context = None
        with self.assertRaises(AssertionError):
            ps.ZeroDurationTask("AZeroDurationTask")

    def test_create_task_zero_duration(self) -> None:
        ps.SchedulingProblem("ProblemWithoutHorizon")
        task = ps.ZeroDurationTask("zdt")
        self.assertIsInstance(task, ps.ZeroDurationTask)

    def test_create_task_fixed_duration(self) -> None:
        new_problem_or_clear()
        task = ps.FixedDurationTask("fdt", 1)
        self.assertIsInstance(task, ps.FixedDurationTask)
        with self.assertRaises(TypeError):
            ps.FixedDurationTask("FloatDuration", 1.0)
        with self.assertRaises(TypeError):
            ps.FixedDurationTask("NullInteger", 0)
        with self.assertRaises(TypeError):
            ps.FixedDurationTask("NegativeInteger", -1)
        with self.assertRaises(TypeError):
            ps.FixedDurationTask("FloatWorkAmount", 2, work_amount=1.0)
        with self.assertRaises(TypeError):
            ps.FixedDurationTask("NegativeWorkAmount", 2, work_amount=-3)

    def test_create_task_variable_duration(self) -> None:
        pb = ps.SchedulingProblem("CreateVariableDurationTask")

        ps.VariableDurationTask("vdt1")
        vdt_2 = ps.VariableDurationTask("vdt2", max_duration=4)
        vdt_3 = ps.VariableDurationTask("vdt3", min_duration=5)

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()

        self.assertTrue(solution)
        self.assertTrue(solution.tasks[vdt_2.name].duration <= 4)
        self.assertTrue(solution.tasks[vdt_3.name].duration >= 5)

    def test_task_types(self) -> None:
        new_problem_or_clear()
        with self.assertRaises(TypeError):
            ps.VariableDurationTask("vdt5", max_duration=4.5)
        with self.assertRaises(TypeError):
            ps.VariableDurationTask("vdt6", max_duration=-1)
        with self.assertRaises(TypeError):
            ps.VariableDurationTask("vdt7", min_duration=-1)
        with self.assertRaises(TypeError):
            ps.VariableDurationTask("vdt8", work_amount=-1)
        with self.assertRaises(TypeError):
            ps.VariableDurationTask("vdt9", work_amount=1.5)
        with self.assertRaises(TypeError):
            ps.VariableDurationTask("vdt10", work_amount=None)

    def test_task_same_names(self) -> None:
        new_problem_or_clear()
        ps.VariableDurationTask("t1")
        with self.assertRaises(ValueError):
            ps.VariableDurationTask("t1")

    def test_eq_overloading(self) -> None:
        new_problem_or_clear()
        task_1 = ps.ZeroDurationTask("task1")
        task_2 = ps.ZeroDurationTask("task2")
        self.assertEqual(task_1, task_1)
        self.assertNotEqual(task_1, task_2)

    def test_redondant_tasks_resources(self) -> None:
        pb = ps.SchedulingProblem("SameNameTasks")
        # we should not be able to add twice the same resource or task
        task_1 = ps.ZeroDurationTask("task1")
        self.assertEqual(list(pb.context.tasks), [task_1])
        self.assertEqual(list(pb.context.tasks), [task_1])
        # do the same for resources
        worker_1 = ps.Worker("Worker1")
        self.assertEqual(list(pb.context.resources), [worker_1])
        self.assertEqual(list(pb.context.resources), [worker_1])

    def test_resource_requirements(self) -> None:
        pb = ps.SchedulingProblem("ResourceRequirements")

        task_1 = ps.FixedDurationTask("task1", duration=3)
        worker_1 = ps.Worker("Worker1")
        worker_2 = ps.Worker("Worker2")
        worker_3 = ps.Worker("Worker3")
        worker_4 = ps.Worker("Worker4")
        task_1.add_required_resource(worker_1)
        task_1.add_required_resource(worker_2)
        task_1.add_required_resources([worker_3, worker_4])
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertEqual(len(solution.tasks["task1"].assigned_resources), 4)

    def test_wrong_assignement(self) -> None:
        new_problem_or_clear()
        task_1 = ps.FixedDurationTask("task1", duration=3)
        worker_1 = ps.Worker("Worker1")
        task_1.add_required_resource(worker_1)
        with self.assertRaises(TypeError):
            task_1.add_required_resource(3)
        with self.assertRaises(TypeError):
            task_1.add_required_resources("a_string")
        with self.assertRaises(ValueError):
            task_1.add_required_resource(worker_1)

    #
    # Single task constraints
    #
    def test_create_task_start_at(self) -> None:
        new_problem_or_clear()
        task = ps.FixedDurationTask("task", 2)
        c = ps.TaskStartAt(task, 1)
        self.assertIsInstance(c, ps.TaskStartAt)
        self.assertEqual(c.value, 1)

    def test_create_task_start_after_strict(self) -> None:
        new_problem_or_clear()
        task = ps.FixedDurationTask("task", 2)
        c = ps.TaskStartAfterStrict(task, 3)
        self.assertIsInstance(c, ps.TaskStartAfterStrict)
        self.assertEqual(c.value, 3)

    def test_create_task_start_after_lax(self) -> None:
        new_problem_or_clear()
        task = ps.FixedDurationTask("task", 2)
        c = ps.TaskStartAfterLax(task, 3)
        self.assertIsInstance(c, ps.TaskStartAfterLax)
        self.assertEqual(c.value, 3)

    def test_create_task_end_at(self) -> None:
        new_problem_or_clear()
        task = ps.FixedDurationTask("task", 2)
        c = ps.TaskEndAt(task, 3)
        self.assertIsInstance(c, ps.TaskEndAt)
        self.assertEqual(c.value, 3)

    def test_create_task_before_strict(self) -> None:
        new_problem_or_clear()
        task = ps.FixedDurationTask("task", 2)
        c = ps.TaskEndBeforeStrict(task, 3)
        self.assertIsInstance(c, ps.TaskEndBeforeStrict)
        self.assertEqual(c.value, 3)

    def test_create_task_before_lax(self) -> None:
        new_problem_or_clear()
        task = ps.FixedDurationTask("task", 2)
        constraint = ps.TaskEndBeforeLax(task, 3)
        self.assertIsInstance(constraint, ps.TaskEndBeforeLax)
        self.assertEqual(constraint.value, 3)

    def test_task_duration_depend_on_start(self) -> None:
        pb = ps.SchedulingProblem("TaskDurationDependsOnStart", horizon=30)

        task_1 = ps.VariableDurationTask("Task1")
        task_2 = ps.VariableDurationTask("Task2")

        pb.add_constraint(ps.TaskStartAt(task_1, 5))
        task_1_cstr = task_1.duration == task_1.start * 3
        pb.add_constraint(task_1_cstr)

        pb.add_constraint(ps.TaskStartAt(task_2, 11))
        task_2_cstr = ps.if_then_else(
            task_2.start < 10, [task_2.duration == 3], [task_2.duration == 1]
        )
        pb.add_constraint(task_2_cstr)

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertEqual(solution.tasks["Task1"].duration, 15)
        self.assertEqual(solution.tasks["Task2"].duration, 1)

    #
    # Two tasks constraints
    #
    def test_create_task_precedence_lax(self) -> None:
        new_problem_or_clear()
        t_1 = ps.FixedDurationTask("t1", duration=2)
        t_2 = ps.FixedDurationTask("t2", duration=3)
        precedence_constraint = ps.TaskPrecedence(t_1, t_2, offset=1, kind="lax")
        self.assertIsInstance(precedence_constraint, ps.TaskPrecedence)

    def test_create_task_precedence_tight(self) -> None:
        new_problem_or_clear()
        t_1 = ps.FixedDurationTask("t1", duration=2)
        t_2 = ps.FixedDurationTask("t2", duration=3)
        precedence_constraint = ps.TaskPrecedence(t_1, t_2, offset=1, kind="tight")
        self.assertIsInstance(precedence_constraint, ps.TaskPrecedence)

    def test_create_task_precedence_strict(self) -> None:
        new_problem_or_clear()
        t_1 = ps.FixedDurationTask("t1", duration=2)
        t_2 = ps.FixedDurationTask("t2", duration=3)
        precedence_constraint = ps.TaskPrecedence(t_1, t_2, offset=1, kind="strict")
        self.assertIsInstance(precedence_constraint, ps.TaskPrecedence)

    def test_create_task_precedence_raise_exception_kind(self) -> None:
        new_problem_or_clear()
        t_1 = ps.FixedDurationTask("t1", duration=2)
        t_2 = ps.FixedDurationTask("t2", duration=3)
        with self.assertRaises(ValueError):
            ps.TaskPrecedence(t_1, t_2, offset=1, kind="foo")

    def test_create_task_precedence_raise_exception_offset_int(self) -> None:
        new_problem_or_clear()
        t_1 = ps.FixedDurationTask("t1", duration=2)
        t_2 = ps.FixedDurationTask("t2", duration=3)
        with self.assertRaises(ValueError):
            ps.TaskPrecedence(t_1, t_2, offset=1.5, kind="lax")  # should be int

    def test_tasks_dont_overlap(self) -> None:
        new_problem_or_clear()
        t_1 = ps.FixedDurationTask("t1", duration=2)
        t_2 = ps.FixedDurationTask("t2", duration=3)
        constraint = ps.TasksDontOverlap(t_1, t_2)
        self.assertIsInstance(constraint, ps.TasksDontOverlap)

    def test_tasks_start_sync(self) -> None:
        new_problem_or_clear()
        t_1 = ps.FixedDurationTask("t1", duration=2)
        t_2 = ps.FixedDurationTask("t2", duration=3)
        constraint = ps.TasksStartSynced(t_1, t_2)
        self.assertIsInstance(constraint, ps.TasksStartSynced)

    def test_tasks_end_sync(self) -> None:
        new_problem_or_clear()
        t_1 = ps.FixedDurationTask("t1", duration=2)
        t_2 = ps.FixedDurationTask("t2", duration=3)
        constraint = ps.TasksEndSynced(t_1, t_2)
        self.assertIsInstance(constraint, ps.TasksEndSynced)

    def test_schedule_n_task_raise_exception(self) -> None:
        new_problem_or_clear()
        with self.assertRaises(TypeError):
            ps.ScheduleNTasksInTimeIntervals(
                "list_of_tasks", 1, "list_of_time_intervals"
            )
        with self.assertRaises(TypeError):
            ps.ScheduleNTasksInTimeIntervals([], 1, "list_of_time_intervals")


if __name__ == "__main__":
    unittest.main()



"""Task constraints and related classes."""

# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import uuid
from typing import Optional

from z3 import And, Bool, Not, BoolRef, Implies, If, Xor, PbEq, PbGe, PbLe

from processscheduler.base import _Constraint

#
# Tasks constraints for two or more classes
#
class TaskPrecedence(_Constraint):
    """Task precedence relation"""

    def __init__(
        self,
        task_before,
        task_after,
        offset=0,
        kind="lax",
        optional: Optional[bool] = False,
    ):
        """kind might be either LAX/STRICT/TIGHT
        Semantics : task after will start at least after offset periods
        task_before is finished.
        LAX constraint: task1_before_end + offset <= task_after_start
        STRICT constraint: task1_before_end + offset < task_after_start
        TIGHT constraint: task1_before_end + offset == task_after_start
        """
        super().__init__(optional)

        if kind not in ["lax", "strict", "tight"]:
            raise ValueError("kind must either be 'lax', 'strict' or 'tight'")

        if not isinstance(offset, int) or offset < 0:
            raise ValueError("offset must be a positive integer")

        self.offset = offset
        self.kind = kind

        if offset > 0:
            lower = task_before.end + offset
        else:
            lower = task_before.end
        upper = task_after.start

        if kind == "lax":
            scheduled_assertion = lower <= upper
        elif kind == "strict":
            scheduled_assertion = lower < upper
        else:  # kind == 'tight':
            scheduled_assertion = lower == upper

        if task_before.optional or task_after.optional:
            # both tasks must be scheduled so that the precedence constraint applies
            self.set_assertions(
                Implies(
                    And(task_before.scheduled, task_after.scheduled),
                    scheduled_assertion,
                )
            )
        else:
            self.set_assertions(scheduled_assertion)


class TasksStartSynced(_Constraint):
    """Two tasks that must start at the same time"""

    def __init__(self, task_1, task_2, optional: Optional[bool] = False) -> None:
        super().__init__(optional)

        scheduled_assertion = task_1.start == task_2.start

        if task_1.optional or task_2.optional:
            # both tasks must be scheduled so that the startsynced constraint applies
            self.set_assertions(
                Implies(And(task_1.scheduled, task_2.scheduled), scheduled_assertion)
            )
        else:
            self.set_assertions(scheduled_assertion)


class TasksEndSynced(_Constraint):
    """Two tasks that must complete at the same time"""

    def __init__(self, task_1, task_2, optional: Optional[bool] = False) -> None:
        super().__init__(optional)

        scheduled_assertion = task_1.end == task_2.end

        if task_1.optional or task_2.optional:
            # both tasks must be scheduled so that the endsynced constraint applies
            self.set_assertions(
                Implies(And(task_1.scheduled, task_2.scheduled), scheduled_assertion)
            )
        else:
            self.set_assertions(scheduled_assertion)


class TasksDontOverlap(_Constraint):
    """two tasks must not overlap, i.e. one needs to be completed before
    the other can be processed"""

    def __init__(self, task_1, task_2, optional: Optional[bool] = False) -> None:
        super().__init__(optional)

        scheduled_assertion = Xor(
            task_2.start >= task_1.end, task_1.start >= task_2.end
        )

        if task_1.optional or task_2.optional:
            # if one task is not scheduledboth tasks must be scheduled so that the not overlap constraint applies
            self.set_assertions(
                Implies(And(task_1.scheduled, task_2.scheduled), scheduled_assertion)
            )
        else:
            self.set_assertions(scheduled_assertion)


#
# Task constraints for one single task
#
class TaskStartAt(_Constraint):
    """One task must start at the desired time"""

    def __init__(self, task, value: int, optional: Optional[bool] = False) -> None:
        super().__init__(optional)
        self.value = value

        scheduled_assertion = task.start == value

        if task.optional:
            self.set_assertions(Implies(task.scheduled, scheduled_assertion))
        else:
            self.set_assertions(scheduled_assertion)


class TaskStartAfterStrict(_Constraint):
    """task.start > value"""

    def __init__(self, task, value: int, optional: Optional[bool] = False) -> None:
        super().__init__(optional)
        self.value = value

        scheduled_assertion = task.start > value

        if task.optional:
            self.set_assertions(Implies(task.scheduled, scheduled_assertion))
        else:
            self.set_assertions(scheduled_assertion)


class TaskStartAfterLax(_Constraint):
    """task.start >= value"""

    def __init__(self, task, value: int, optional: Optional[bool] = False) -> None:
        super().__init__(optional)
        self.value = value

        scheduled_assertion = task.start >= value

        if task.optional:
            self.set_assertions(Implies(task.scheduled, scheduled_assertion))
        else:
            self.set_assertions(scheduled_assertion)


class TaskEndAt(_Constraint):
    """On task must complete at the desired time"""

    def __init__(self, task, value: int, optional: Optional[bool] = False) -> None:
        super().__init__(optional)
        self.value = value

        scheduled_assertion = task.end == value

        if task.optional:
            self.set_assertions(Implies(task.scheduled, scheduled_assertion))
        else:
            self.set_assertions(scheduled_assertion)


class TaskEndBeforeStrict(_Constraint):
    """task.end < value"""

    def __init__(self, task, value: int, optional: Optional[bool] = False) -> None:
        super().__init__(optional)
        self.value = value

        scheduled_assertion = task.end < value

        if task.optional:
            self.set_assertions(Implies(task.scheduled, scheduled_assertion))
        else:
            self.set_assertions(scheduled_assertion)


class TaskEndBeforeLax(_Constraint):
    """task.end <= value"""

    def __init__(self, task, value: int, optional: Optional[bool] = False) -> None:
        super().__init__(optional)
        self.value = value

        scheduled_assertion = task.end <= value

        if task.optional:
            self.set_assertions(Implies(task.scheduled, scheduled_assertion))
        else:
            self.set_assertions(scheduled_assertion)


#
# Optional classes only constraints
#
class OptionalTaskConditionSchedule(_Constraint):
    """An optional task that is scheduled only if a condition is fulfilled."""

    def __init__(
        self, task, condition: BoolRef, optional: Optional[bool] = False
    ) -> None:
        super().__init__(optional)

        if not task.optional:
            raise TypeError("Task %s must be optional." % task.name)

        self.set_assertions(
            If(condition, task.scheduled == True, task.scheduled == False)
        )


class OptionalTasksDependency(_Constraint):
    """task_2 is scheduled if and only if task_1 is scheduled"""

    def __init__(self, task_1, task_2, optional: Optional[bool] = False) -> None:
        super().__init__(optional)

        if not task_2.optional:
            raise TypeError("Task %s must be optional." % task_2.name)

        self.set_assertions(task_1.scheduled == task_2.scheduled)


class ForceScheduleNOptionalTasks(_Constraint):
    """Given a set of m different optional tasks, force the solver to schedule
    at at least/at most/exactly n tasks, with 0 < n <= m."""

    def __init__(
        self,
        list_of_optional_tasks,
        nb_tasks_to_schedule: Optional[int] = 1,
        kind: Optional[str] = "exact",
        optional: Optional[bool] = False,
    ) -> None:
        super().__init__(optional)

        problem_function = {"min": PbGe, "max": PbLe, "exact": PbEq}

        # first check that all tasks from the list_of_optional_tasks are
        # actually optional
        for task in list_of_optional_tasks:
            if not task.optional:
                raise TypeError(
                    "This class %s must excplicitely be set as optional." % task.name
                )
        # all scheduled variables to take into account
        sched_vars = []
        for task in list_of_optional_tasks:
            sched_vars.append(task.scheduled)

        asst = problem_function[kind](
            [(scheduled, True) for scheduled in sched_vars], nb_tasks_to_schedule
        )
        self.set_assertions(asst)


class ScheduleNTasksInTimeIntervals(_Constraint):
    """Given a set of m different tasks, and a list of time intervals, schedule N tasks among m
    in this time interval"""

    def __init__(
        self,
        list_of_tasks,
        nb_tasks_to_schedule,
        list_of_time_intervals,
        kind: Optional[str] = "exact",
        optional: Optional[bool] = False,
    ) -> None:
        super().__init__(optional)

        problem_function = {"min": PbGe, "max": PbLe, "exact": PbEq}

        # first check that all tasks from the list_of_optional_tasks are
        # actually optional
        if not isinstance(list_of_tasks, list):
            raise TypeError("list_of_task must be a list")

        if not isinstance(list_of_time_intervals, list):
            raise TypeError("list_of_time_intervals must be a list of list")

        # count the number of tasks that re scheduled in this time interval
        all_bools = []
        for task in list_of_tasks:
            # for this task, the logic expression is that any of its start or end must be
            # between two consecutive intervals
            bools_for_this_task = []
            for time_interval in list_of_time_intervals:
                task_in_time_interval = Bool(
                    "InTimeIntervalTask_%s_%i" % (task.name, uuid.uuid4().int)
                )
                lower_bound, upper_bound = time_interval
                cstrs = [
                    task.start >= lower_bound,
                    task.end <= upper_bound,
                    Not(
                        And(task.start < lower_bound, task.end > lower_bound)
                    ),  # overlap at start
                    Not(
                        And(task.start < upper_bound, task.end > upper_bound)
                    ),  # overlap at end
                    Not(And(task.start < lower_bound, task.end > upper_bound)),
                ]  # full overlap
                asst = Implies(task_in_time_interval, And(cstrs))
                self.set_assertions(asst)
                bools_for_this_task.append(task_in_time_interval)
            # only one maximum bool to True from the previous possibilities
            asst_tsk = PbLe([(scheduled, True) for scheduled in bools_for_this_task], 1)
            self.set_assertions(asst_tsk)
            all_bools.extend(bools_for_this_task)

        # we also have to exclude all the other cases, where start or end can be between two intervals
        # then set the constraint for the number of tasks to schedule
        asst_pb = problem_function[kind](
            [(scheduled, True) for scheduled in all_bools], nb_tasks_to_schedule
        )
        self.set_assertions(asst_pb)



# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import unittest

import processscheduler as ps


class TestIndicator(unittest.TestCase):
    def test_indicator_flowtime(self) -> None:
        problem = ps.SchedulingProblem("IndicatorFlowTime", horizon=2)
        t_1 = ps.FixedDurationTask("t1", 2)
        t_2 = ps.FixedDurationTask("t2", 2)

        i_1 = ps.Indicator("FlowTime", t_1.end + t_2.end)

        solution = ps.SchedulingSolver(problem).solve()

        self.assertTrue(solution)
        self.assertEqual(solution.indicators[i_1.name], 4)

    def test_resource_utilization_indicator_1(self) -> None:
        problem = ps.SchedulingProblem("IndicatorUtilization1", horizon=10)
        t_1 = ps.FixedDurationTask("T1", duration=5)
        worker_1 = ps.Worker("Worker1")
        t_1.add_required_resource(worker_1)
        utilization_ind = problem.add_indicator_resource_utilization(worker_1)

        solution = ps.SchedulingSolver(problem).solve()

        self.assertTrue(solution)
        self.assertEqual(solution.indicators[utilization_ind.name], 50)

    def test_resource_utilization_indicator_2(self) -> None:
        """Two tasks, two workers."""
        problem = ps.SchedulingProblem("IndicatorUtilization2", horizon=10)

        t_1 = ps.FixedDurationTask("T1", duration=5)
        t_2 = ps.FixedDurationTask("T2", duration=5)

        worker_1 = ps.Worker("Worker1")
        worker_2 = ps.Worker("Worker2")

        t_1.add_required_resource(worker_1)
        t_2.add_required_resource(ps.SelectWorkers([worker_1, worker_2]))

        utilization_res_1 = problem.add_indicator_resource_utilization(worker_1)
        utilization_res_2 = problem.add_indicator_resource_utilization(worker_2)

        solution = ps.SchedulingSolver(problem).solve()

        self.assertTrue(solution)
        result_res_1 = solution.indicators[utilization_res_1.name]
        result_res_2 = solution.indicators[utilization_res_2.name]

        # sum should be 100
        self.assertEqual(result_res_1 + result_res_2, 100)

    def test_resource_utilization_indicator_3(self) -> None:
        """Same as above, but both workers are selectable. Force one with resource
        utilization maximization objective."""
        problem = ps.SchedulingProblem("IndicatorUtilization3", horizon=10)

        t_1 = ps.FixedDurationTask("T1", duration=5)
        t_2 = ps.FixedDurationTask("T2", duration=5)

        worker_1 = ps.Worker("Worker1")
        worker_2 = ps.Worker("Worker2")

        t_1.add_required_resource(ps.SelectWorkers([worker_1, worker_2]))
        t_2.add_required_resource(ps.SelectWorkers([worker_1, worker_2]))

        utilization_res_1 = problem.add_indicator_resource_utilization(worker_1)
        utilization_res_2 = problem.add_indicator_resource_utilization(worker_2)

        ps.MaximizeObjective("MaximieResource1Utilization", utilization_res_1)

        solution = ps.SchedulingSolver(problem).solve()

        self.assertTrue(solution)
        self.assertEqual(solution.indicators[utilization_res_1.name], 100)
        self.assertEqual(solution.indicators[utilization_res_2.name], 0)

    def test_resource_utilization_indicator_4(self) -> None:
        """20 optional tasks, one worker. Force resource utilization maximization objective."""
        problem = ps.SchedulingProblem("IndicatorUtilization4", horizon=20)

        worker = ps.Worker("Worker")

        for i in range(20):
            t = ps.FixedDurationTask(f"T{i+1}", duration=1, optional=True)
            t.add_required_resource(worker)

        utilization_res = problem.add_indicator_resource_utilization(worker)
        problem.maximize_indicator(utilization_res)

        solution = ps.SchedulingSolver(problem).solve()

        self.assertTrue(solution)
        self.assertEqual(solution.indicators[utilization_res.name], 100)

    # TODO: test commented out because it sometimes fails on Azure cloud
    # def test_resource_utilization_indicator_5(self) -> None:
    #     """Same input data than previous tests, but we dont use
    #     an optimisation solver, the objective of 100% is set by an
    #     additional constraint. This should be **much faster**."""
    #     problem = ps.SchedulingProblem('IndicatorUtilization5', horizon = 20)

    #     worker = ps.Worker('Worker')

    #     for i in range(40):
    #         t = ps.FixedDurationTask(f'T{i+1}', duration = 1, optional = True)
    #         t.add_required_resource(worker)

    #     utilization_res = problem.add_indicator_resource_utilization(worker)

    #     problem.add_constraint(utilization_res.indicator_variable == 100)

    #     solution = ps.SchedulingSolver(problem).solve()

    #     self.assertTrue(solution)
    #     self.assertEqual(solution.indicators[utilization_res.name], 100)

    def test_optimize_indicator(self) -> None:
        problem = ps.SchedulingProblem("SolvePriorities", horizon=10)
        task_1 = ps.FixedDurationTask("task1", duration=2, priority=1)
        task_2 = ps.FixedDurationTask("task2", duration=2, priority=10, optional=True)
        task_3 = ps.FixedDurationTask("task3", duration=2, priority=100, optional=True)

        worker = ps.Worker("AWorker")
        task_1.add_required_resource(worker)
        task_2.add_required_resource(worker)
        task_3.add_required_resource(worker)

        problem.add_objective_resource_utilization(worker)
        problem.add_objective_priorities()

        solver = ps.SchedulingSolver(problem)
        solution = solver.solve()
        self.assertTrue(solution)

    def test_incremental_optimizer_1(self) -> None:
        problem = ps.SchedulingProblem("IncrementalOptimizer1", horizon=100)
        task_1 = ps.FixedDurationTask("task1", duration=2)

        task_1_start_ind = ps.Indicator("Task1Start", task_1.start)
        problem.maximize_indicator(task_1_start_ind)

        solver = ps.SchedulingSolver(problem)
        solution = solver.solve()

        self.assertTrue(solution)
        self.assertEqual(solution.indicators[task_1_start_ind.name], 98)

    def test_resource_utilization_maximization_incremental_1(self) -> None:
        """Same as above, but both workers are selectable. Force one with resource
        utilization maximization objective."""
        problem = ps.SchedulingProblem("IndicatorMaximizeIncremental", horizon=10)

        t_1 = ps.FixedDurationTask("T1", duration=5)
        t_2 = ps.FixedDurationTask("T2", duration=5)

        worker_1 = ps.Worker("Worker1")
        worker_2 = ps.Worker("Worker2")

        t_1.add_required_resource(ps.SelectWorkers([worker_1, worker_2]))
        t_2.add_required_resource(ps.SelectWorkers([worker_1, worker_2]))

        utilization_res_1 = problem.add_indicator_resource_utilization(worker_1)
        utilization_res_2 = problem.add_indicator_resource_utilization(worker_2)

        problem.maximize_indicator(utilization_res_1)
        solver = ps.SchedulingSolver(problem)

        solution = solver.solve()

        self.assertTrue(solution)
        self.assertEqual(solution.indicators[utilization_res_1.name], 100)
        self.assertEqual(solution.indicators[utilization_res_2.name], 0)

    def get_single_resource_utilization_problem(self, problem_name):
        problem = ps.SchedulingProblem("IndicatorFlowtimeSingleResource1", horizon=50)

        dur1 = 5
        dur2 = 5
        dur3 = 4
        dur4 = 3
        dur5 = 2

        t_1 = ps.FixedDurationTask("T1", duration=dur1)
        t_2 = ps.FixedDurationTask("T2", duration=dur2)
        t_3 = ps.FixedDurationTask("T3", duration=dur3)
        t_4 = ps.FixedDurationTask("T4", duration=dur4)
        t_5 = ps.FixedDurationTask("T5", duration=dur5)
        worker_1 = ps.Worker("Worker1")

        t_1.add_required_resource(worker_1)
        t_2.add_required_resource(worker_1)
        t_3.add_required_resource(worker_1)
        t_4.add_required_resource(worker_1)
        t_5.add_required_resource(worker_1)

        problem.add_constraint(ps.TaskEndBeforeLax(t_3, 35))
        problem.add_constraint(ps.TaskEndBeforeLax(t_2, 35))
        problem.add_constraint(ps.TaskEndBeforeLax(t_1, 35))
        problem.add_constraint(ps.TaskEndBeforeLax(t_4, 35))
        problem.add_constraint(ps.TaskEndBeforeLax(t_5, 35))

        problem.add_constraint(ps.TaskStartAfterLax(t_3, 10))
        problem.add_constraint(ps.TaskStartAfterLax(t_2, 10))
        problem.add_constraint(ps.TaskStartAfterLax(t_1, 10))
        problem.add_constraint(ps.TaskStartAfterLax(t_4, 10))
        problem.add_constraint(ps.TaskStartAfterLax(t_5, 10))

        return problem, worker_1, dur1 + dur2 + dur3 + dur4 + dur5

    def test_indicator_flowtime_single_resource_1(self) -> None:
        problem, worker_1, _ = self.get_single_resource_utilization_problem(
            "IndicatorFlowtimeSingleResource1"
        )
        # there should not be any task scheduled in this time period
        problem.add_objective_flowtime_single_resource(worker_1, time_interval=[40, 50])
        solver = ps.SchedulingSolver(problem)

        solution = solver.solve()
        self.assertTrue(solution)
        # the flowtime should be 0
        self.assertEqual(solution.indicators["FlowTime(Worker1:40:50)"], 0)

    def test_indicator_flowtime_single_resource_2(self) -> None:
        # same as before, but the time interval contains all the tasks
        problem, worker_1, sum_durations = self.get_single_resource_utilization_problem(
            "IndicatorFlowtimeSingleResource2"
        )
        problem.add_objective_flowtime_single_resource(worker_1, time_interval=[10, 40])
        solver = ps.SchedulingSolver(problem)

        solution = solver.solve()
        self.assertTrue(solution)
        self.assertEqual(solution.indicators["FlowTime(Worker1:10:40)"], sum_durations)

    def test_indicator_flowtime_single_resource_3(self) -> None:
        # same as before, but the time interval contains no task
        problem, worker_1, _ = self.get_single_resource_utilization_problem(
            "IndicatorFlowtimeSingleResource3"
        )
        problem.add_objective_flowtime_single_resource(worker_1, time_interval=[5, 9])
        solver = ps.SchedulingSolver(problem)

        solution = solver.solve()
        self.assertTrue(solution)
        print(solution)
        self.assertEqual(solution.indicators["FlowTime(Worker1:5:9)"], 0)

    def test_indicator_flowtime_single_resource_4(self) -> None:
        # without any time_interval provided, should use the whole range [0, horizon]
        problem, worker_1, sum_durations = self.get_single_resource_utilization_problem(
            "IndicatorFlowtimeSingleResource4"
        )
        problem.add_objective_flowtime_single_resource(worker_1)
        solver = ps.SchedulingSolver(problem)

        solution = solver.solve()
        print(solution)
        self.assertTrue(solution)
        self.assertEqual(
            solution.indicators["FlowTime(Worker1:0:horizon)"], sum_durations
        )

    @staticmethod
    def get_single_resource_utilization_problem_2(time_intervals):
        horizon = time_intervals[-1][-1] + 3
        nb_tasks = [5 for interval in time_intervals if interval[1] - interval[0] > 3]
        problem = ps.SchedulingProblem(
            "IndicatorFlowtimeSingleResource", horizon=horizon
        )
        worker_1 = ps.Worker("Worker1")

        tasks: list[ps.FixedDurationTask] = []
        for interval, nb_tasks_i in zip(time_intervals, nb_tasks):
            for _ in range(nb_tasks_i):
                tasks.append(ps.FixedDurationTask("T%s" % len(tasks), duration=1))
                tasks[-1].add_required_resource(worker_1)
                problem.add_constraint(ps.TaskStartAfterLax(tasks[-1], interval[0]))
                problem.add_constraint(ps.TaskEndBeforeLax(tasks[-1], interval[1]))
        return problem, worker_1, len(tasks)

    @staticmethod
    def get_sum_flowtime(solution) -> int:
        return sum(
            [
                solution.indicators[indicator_id]
                for indicator_id in solution.indicators
                if "FlowTime" in indicator_id
            ]
        )

    def test_indicator_flowtime_single_resource_5(self) -> None:
        # 2 time interval objectives
        time_intervals = [(11, 20), (21, 34)]
        (
            problem,
            worker_1,
            sum_durations,
        ) = self.get_single_resource_utilization_problem_2(time_intervals)
        for interval in time_intervals:
            problem.add_objective_flowtime_single_resource(
                worker_1, time_interval=interval
            )

        solver = ps.SchedulingSolver(problem)
        solution = solver.solve()

        self.assertTrue(solution)
        self.assertEqual(sum_durations, self.get_sum_flowtime(solution))

    def test_indicator_flowtime_single_resource_6(self) -> None:
        # Mutliple time intervals (Currently fails for nb_time_intervals > 2, gantt to check total_flowtime is correct)
        nb_time_intervals = 7
        time_interval_length = 13  # always > 5
        horizon = nb_time_intervals * time_interval_length
        time_intervals = [
            (i, i + time_interval_length)
            for i in range(0, horizon, time_interval_length)
        ]

        (
            problem,
            worker_1,
            sum_durations,
        ) = self.get_single_resource_utilization_problem_2(time_intervals)
        for interval in time_intervals:
            problem.add_objective_flowtime_single_resource(
                worker_1, time_interval=interval
            )

        solver = ps.SchedulingSolver(problem)

        solution = solver.solve()

        self.assertTrue(solution)
        self.assertEqual(sum_durations, self.get_sum_flowtime(solution))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

from processscheduler.base import _NamedUIDObject, is_positive_integer


class _Cost(_NamedUIDObject):
    """The base class for cost definition, to be assigned to a resource instance"""

    def __init__(self):
        super().__init__("")
        self.f = lambda x: 0  # by default

    def plot(self, interval, show_plot=True) -> None:
        """Plot the cost curve using matplotlib."""
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            raise ModuleNotFoundError("matplotlib is not installed.")

        try:
            import numpy as np
        except ImportError:
            raise ModuleNotFoundError("numpy is not installed.")

        lower_bound, upper_bound = interval
        x = np.linspace(lower_bound, upper_bound, 1000)
        y = []
        for x_ in x:
            y.append(self.f(x_))

        plt.plot(x, y, label="Cost function")

        plt.legend()
        plt.grid(True)
        plt.xlabel("x")
        plt.ylabel("y")

        if show_plot:
            plt.show()


class ConstantCostPerPeriod(_Cost):
    def __init__(self, value: int) -> None:
        super().__init__()
        if not is_positive_integer(value):
            raise ValueError("the cost per period must be a positive integer")
        self.value = value
        self.f = lambda x: value

    def __call__(self, value):
        """compute the value of the cost function for a given value"""
        return self.f(value)


class PolynomialCostFunction(_Cost):
    """A function of time under a polynomail form."""

    def __init__(self, function: callable) -> None:
        super().__init__()
        if not callable(function):
            raise TypeError("function must be a callable")
        self.f = function

    def __call__(self, value):
        """compute the value of the cost function for a given value"""
        return self.f(value)



# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import unittest

import processscheduler as ps


class TestResourceUnavailable(unittest.TestCase):
    def test_resource_unavailable_1(self) -> None:
        pb = ps.SchedulingProblem("ResourceUnavailable1", horizon=10)
        task_1 = ps.FixedDurationTask("task1", duration=3)
        worker_1 = ps.Worker("Worker1")
        task_1.add_required_resource(worker_1)
        c1 = ps.ResourceUnavailable(worker_1, [(1, 3), (6, 8)])
        pb.add_constraint(c1)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertEqual(solution.tasks[task_1.name].start, 3)
        self.assertEqual(solution.tasks[task_1.name].end, 6)

    def test_resource_unavailable_2(self) -> None:
        pb = ps.SchedulingProblem("ResourceUnavailable2", horizon=10)
        task_1 = ps.FixedDurationTask("task1", duration=3)
        worker_1 = ps.Worker("Worker1")
        task_1.add_required_resource(worker_1)
        # difference with the first one: build 2 constraints
        # merged using a and_
        c1 = ps.ResourceUnavailable(worker_1, [(1, 3)])
        c2 = ps.ResourceUnavailable(worker_1, [(6, 8)])
        pb.add_constraint(ps.and_([c1, c2]))
        # that should not change the problem solution
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertEqual(solution.tasks[task_1.name].start, 3)
        self.assertEqual(solution.tasks[task_1.name].end, 6)

    def test_resource_unavailable_3(self) -> None:
        pb = ps.SchedulingProblem("ResourceUnavailable3", horizon=10)
        task_1 = ps.FixedDurationTask("task1", duration=3)
        worker_1 = ps.Worker("Worker1")
        task_1.add_required_resource(worker_1)
        # difference with the previous ones: too much unavailability,
        # so possible solution
        # merged using a and_
        c1 = ps.ResourceUnavailable(worker_1, [(1, 3)])
        c2 = ps.ResourceUnavailable(worker_1, [(5, 8)])
        pb.add_constraint(ps.and_([c1, c2]))
        # that should not change the problem solution
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertFalse(solution)

    def test_cumulative_4(self):
        pb_bs = ps.SchedulingProblem("ResourceUnavailableCumulative1", 10)
        # tasks
        t1 = ps.FixedDurationTask("T1", duration=2)
        t2 = ps.FixedDurationTask("T2", duration=2)
        t3 = ps.FixedDurationTask("T3", duration=2)

        # workers
        r1 = ps.CumulativeWorker("Machine1", size=3)

        # resource assignment
        t1.add_required_resource(r1)
        t2.add_required_resource(r1)
        t3.add_required_resource(r1)

        c1 = ps.ResourceUnavailable(r1, [(1, 10)])
        pb_bs.add_constraint(c1)

        # plot solution
        solver = ps.SchedulingSolver(pb_bs)  # , debug=True)
        solution = solver.solve()
        self.assertFalse(solution)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import unittest

import processscheduler as ps


class TestWorkLoad(unittest.TestCase):
    def test_resource_work_load_basic(self) -> None:
        pb = ps.SchedulingProblem("ResourceWorkLoadBasic", horizon=12)
        task_1 = ps.FixedDurationTask("task1", duration=8)

        worker_1 = ps.Worker("Worker1")
        task_1.add_required_resource(worker_1)

        c1 = ps.WorkLoad(worker_1, {(0, 6): 2})
        pb.add_constraint(c1)

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        # the only possible solution is that the task is scheduled form 4 to 12
        self.assertEqual(solution.tasks[task_1.name].start, 4)
        self.assertEqual(solution.tasks[task_1.name].end, 12)

    def test_resource_work_load_eq(self) -> None:
        pb = ps.SchedulingProblem("ResourceWorkLoadEq", horizon=12)
        task_1 = ps.FixedDurationTask("task1", duration=7)

        worker_1 = ps.Worker("Worker1")
        task_1.add_required_resource(worker_1)

        c1 = ps.WorkLoad(worker_1, {(0, 6): 3}, kind="exact")
        pb.add_constraint(c1)
        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        # the only possible solution is that the task is scheduled form 3 to 9
        self.assertEqual(solution.tasks[task_1.name].start, 3)
        self.assertEqual(solution.tasks[task_1.name].end, 10)

    def test_resource_work_load_min(self) -> None:
        pb = ps.SchedulingProblem("ResourceWorkLoadMin", horizon=20)
        task_1 = ps.FixedDurationTask("task1", duration=7)

        worker_1 = ps.Worker("Worker1")
        task_1.add_required_resource(worker_1)

        c1 = ps.WorkLoad(worker_1, {(6, 8): 2}, kind="min")
        pb.add_constraint(c1)

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()

        self.assertTrue(solution)
        # the only possible solution is that the task is scheduled form 3 to 9
        self.assertTrue(solution.tasks[task_1.name].start <= 6)
        self.assertTrue(solution.tasks[task_1.name].end >= 8)

    def test_resource_work_load_3(self) -> None:
        # same problem, but we force two tasks to be scheduled at start and end
        # of the planning
        pb = ps.SchedulingProblem("ResourceWorkLoadUnavailability", horizon=12)
        task_1 = ps.FixedDurationTask("task1", duration=4)
        task_2 = ps.FixedDurationTask("task2", duration=4)
        worker_1 = ps.Worker("Worker1")
        task_1.add_required_resource(worker_1)
        task_2.add_required_resource(worker_1)

        c1 = ps.WorkLoad(worker_1, {(4, 8): 0})
        pb.add_constraint(c1)

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        # there should be one task from 0 to 4 and one task from 8 to 12.
        self.assertTrue(
            solution.tasks[task_1.name].start == 0
            or solution.tasks[task_2.name].start == 0
        )
        self.assertTrue(
            solution.tasks[task_1.name].start == 8
            or solution.tasks[task_2.name].start == 8
        )

    def test_resource_work_load_exception(self) -> None:
        ps.SchedulingProblem("ResourceWorkLoadException", horizon=12)

        worker_1 = ps.Worker("Worker1")

        with self.assertRaises(ValueError):
            ps.WorkLoad(worker_1, {(0, 6): 2}, kind="foo")

    def test_selectworker_work_load_1(self) -> None:
        pb = ps.SchedulingProblem("SelectWorkerWorkLoad1", horizon=12)

        worker_1 = ps.Worker("Worker1")
        worker_2 = ps.Worker("Worker2")

        task1 = ps.FixedDurationTask("Task1", duration=10)

        task1.add_required_resource(ps.SelectWorkers([worker_1, worker_2], 1, "min"))
        # the workload on worker_1 forces 0 between 4 and 8
        # then the worker_1 can not be assigned
        c1 = ps.WorkLoad(worker_1, {(4, 8): 0})
        pb.add_constraint(c1)

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        print(solution)
        self.assertTrue(solution)
        self.assertTrue(
            solution.tasks[task1.name].assigned_resources == [worker_2.name]
        )

    def test_multiple_workers_work_load_1(self) -> None:
        pb = ps.SchedulingProblem("MultipleWorkersWorkLoad1", horizon=12)

        worker_1 = ps.Worker("Worker1")
        worker_2 = ps.Worker("Worker2")

        task1 = ps.FixedDurationTask("Task1", duration=10)
        task1.add_required_resources([worker_1, worker_2])

        c1 = ps.WorkLoad(worker_1, {(0, 4): 2})
        pb.add_constraint(c1)
        c2 = ps.WorkLoad(worker_2, {(2, 6): 2}, kind="min")
        pb.add_constraint(c2)

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertTrue(solution)
        self.assertTrue(solution.tasks[task1.name].start == 2)

    def test_multiple_workers_work_load_2(self) -> None:
        ### the same as above but changing 'min' to 'exact', there's no solution
        pb = ps.SchedulingProblem("MultipleWorkersWorkLoad2", horizon=12)

        worker_1 = ps.Worker("Worker1")
        worker_2 = ps.Worker("Worker2")

        task1 = ps.FixedDurationTask("Task1", duration=10)
        task1.add_required_resources([worker_1, worker_2])

        c1 = ps.WorkLoad(worker_1, {(0, 4): 2})
        pb.add_constraint(c1)
        c2 = ps.WorkLoad(worker_2, {(2, 6): 2}, kind="exact")
        pb.add_constraint(c2)

        solver = ps.SchedulingSolver(pb)
        solution = solver.solve()
        self.assertFalse(solution)


if __name__ == "__main__":
    unittest.main()
